# !/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.18 18:00:00                  #
# ================================================== #

from pygpt_net.core.dispatcher import Event


class Mode:
    def __init__(self, window=None):
        """
        UI mode switch controller

        :param window: Window instance
        """
        self.window = window

    def update(self):
        """Update mode, model, preset and rest of the toolbox"""

        mode = self.window.core.config.data['mode']

        # presets
        if mode != "assistant":
            self.window.ui.nodes['presets.widget'].setVisible(True)
        else:
            self.window.ui.nodes['presets.widget'].setVisible(False)

        # presets: clear
        if mode in ["img", "llama_index", "assistant"]:
            self.window.ui.nodes['preset.clear'].setVisible(False)
        else:
            self.window.ui.nodes['preset.clear'].setVisible(True)

        # presets: use
        if mode == 'img':
            self.window.ui.nodes['preset.use'].setVisible(True)
        else:
            self.window.ui.nodes['preset.use'].setVisible(False)

        # img options
        if mode == "img":
            self.window.ui.nodes['dalle.options'].setVisible(True)
        else:
            self.window.ui.nodes['dalle.options'].setVisible(False)

        # agent options
        if mode == "agent":
            self.window.ui.nodes['agent.options'].setVisible(True)
        else:
            self.window.ui.nodes['agent.options'].setVisible(False)

        # assistants list
        if mode == "assistant":
            self.window.ui.nodes['assistants.widget'].setVisible(True)
        else:
            self.window.ui.nodes['assistants.widget'].setVisible(False)

        # indexes list
        if mode == "llama_index":
            self.window.ui.nodes['indexes.widget'].setVisible(True)
            self.window.ui.nodes['idx.options'].setVisible(True)
        else:
            self.window.ui.nodes['indexes.widget'].setVisible(False)
            self.window.ui.nodes['idx.options'].setVisible(False)

        # stream mode
        if mode in ["img", "assistant"]:
            self.window.ui.nodes['input.stream'].setVisible(False)
        else:
            self.window.ui.nodes['input.stream'].setVisible(True)

        # vision
        show = self.is_vision(mode)
        self.window.ui.nodes['vision.capture.options'].setVisible(show)
        self.window.ui.nodes['attachments.capture_clear'].setVisible(show)

        # attachments
        show = self.are_attachments(mode)
        self.window.ui.tabs['input'].setTabVisible(1, show)  # attachments

        # uploaded files
        if mode == "assistant":
            self.window.ui.tabs['input'].setTabVisible(2, True)
        else:
            self.window.ui.tabs['input'].setTabVisible(2, False)

        # toggle chat footer
        if not self.is_chat_tab():
            self.hide_chat_footer()
        else:
            self.show_chat_footer()

    def is_vision(self, mode: str) -> bool:
        """Check if vision is allowed"""
        if mode == "vision":
            return True

        allowed = self.window.controller.painter.is_active()
        if allowed:
            return True

        if mode in ["img", "assistant"]:
            return False

        # event: UI: vision
        value = False
        event = Event(Event.UI_VISION, {
            'mode': mode,
            'value': value,
        })
        self.window.core.dispatcher.dispatch(event)
        return event.data['value']

    def are_attachments(self, mode: str) -> bool:
        """Check if attachments are allowed"""
        if mode in ["vision", "assistant"]:
            return True

        if mode == "img":
            return False

        # event: UI: attachments
        value = False
        event = Event(Event.UI_ATTACHMENTS, {
            'mode': mode,
            'value': value,
        })
        self.window.core.dispatcher.dispatch(event)
        return event.data['value']

    def is_chat_tab(self) -> bool:
        """Check if current tab is chat"""
        return self.window.controller.ui.current_tab == self.window.controller.ui.tab_idx['chat']

    def show_chat_footer(self):
        """Show chat footer"""
        self.window.ui.nodes['chat.footer'].setVisible(True)

    def hide_chat_footer(self):
        """Hide chat footer"""
        self.window.ui.nodes['chat.footer'].setVisible(False)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.25 22:00:00                  #
# ================================================== #

import json
import ast
from PySide6.QtCore import Qt
from PySide6.QtGui import QCursor
from PySide6.QtWidgets import QAbstractItemView, QApplication, QMenu, \
    QSizePolicy, QTableView, QHeaderView


class DebugList(QTableView):
    NAME = range(2)  # list of columns

    def __init__(self, window=None):
        """
        Debug list

        :param window: Window instance
        """
        super(DebugList, self).__init__(window)
        self.window = window
        self.viewer = None
        self.viewer_index = None
        self.viewer_current = None
        self.selection = None
        self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.verticalHeader().setVisible(False)
        self.setWordWrap(True)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.clicked.connect(self.onItemClicked)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(
            lambda: self.create_context_menu(self))

    def on_data_begin(self):
        """
        On data update begin
        """
        self.adjustColumns()

    def on_data_end(self):
        """
        On data update end
        """
        # update data viewer
        if self.viewer_index:
            new = self.model().data(self.viewer_index)
            if self.viewer_current != new:
                self.viewer.setPlainText(self.parse_view(new))
                self.viewer_current = new

    def parse_view(self, data):
        """
        Parse view, convert data to formatted JSON if needed

        :param data: Data
        """
        try:
            if data.startswith("{") or data.startswith("["):
                data = json.dumps(ast.literal_eval(data), indent=4)
        except Exception as e:
            pass
        return data

    def adjustColumns(self):
        """Adjust columns width"""
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)

    def create_context_menu(self, parent):
        """
        Create context menu

        :param parent: Parent
        """
        def copy_to_clipboard():
            index = parent.currentIndex()
            if index.isValid():
                value = index.sibling(index.row(), 1).data(
                    Qt.DisplayRole)
                QApplication.clipboard().setText(value)

        menu = QMenu()
        copy_action = menu.addAction("Copy value to clipboard")
        copy_action.triggered.connect(copy_to_clipboard)
        menu.exec_(QCursor.pos())

    def mousePressEvent(self, event):
        """
        Mouse press event

        :param event: Event
        """
        index = self.indexAt(event.pos())
        if not index.isValid():
            return
        super(DebugList, self).mousePressEvent(event)

    def onItemClicked(self, index):
        """
        On item clicked

        :param index: Index
        """
        # update data viewer
        data = self.model().data(index)
        self.viewer.setPlainText(self.parse_view(data))
        self.viewer_index = index
        self.viewer_current = data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.25 21:00:00                  #
# ================================================== #

from PySide6.QtGui import QStandardItemModel
from PySide6.QtWidgets import QVBoxLayout, QLabel, QPushButton, QWidget
from datetime import datetime, timedelta

from pygpt_net.ui.widget.lists.context import ContextList
from pygpt_net.ui.widget.textarea.search_input import CtxSearchInput
from pygpt_net.utils import trans


class SearchInput:
    def __init__(self, window=None):
        """
        Context search input

        :param window: Window instance
        """
        self.window = window

    def setup(self) -> QWidget:
        """
        Setup search input

        :return: QWidget
        """
        self.window.ui.nodes['ctx.search'] = CtxSearchInput(self.window)

        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.nodes['ctx.search'])
        layout.setContentsMargins(0, 0, 0, 0)

        widget = QWidget()
        widget.setLayout(layout)
        widget.setContentsMargins(0, 0, 0, 0)

        return widget



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.26 20:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QPushButton, QRadioButton, QCheckBox, QWidget

from pygpt_net.ui.layout.chat.attachments import Attachments
from pygpt_net.ui.layout.chat.attachments_uploaded import AttachmentsUploaded
from pygpt_net.ui.layout.status import Status
from pygpt_net.ui.widget.audio.input import AudioInput
from pygpt_net.ui.widget.audio.input_button import AudioInputButton
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.ui.widget.tabs.Input import InputTabs
from pygpt_net.ui.widget.textarea.input import ChatInput
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class Input:
    def __init__(self, window=None):
        """
        Input UI

        :param window: Window instance
        """
        self.window = window
        self.status = Status(window)
        self.attachments = Attachments(window)
        self.attachments_uploaded = AttachmentsUploaded(window)

        # min height
        self.min_height_files_tab = 120
        self.min_height_input_tab = 80
        self.min_height_input = 50

    def setup(self) -> QWidget:
        """
        Setup input

        :return: QWidget
        """
        input = self.setup_input()
        files = self.setup_attachments()
        files_uploaded = self.setup_attachments_uploaded()

        # tabs
        self.window.ui.tabs['input'] = InputTabs(self.window)
        self.window.ui.tabs['input'].setMinimumHeight(self.min_height_input_tab)
        self.window.ui.tabs['input'].addTab(input, trans('input.tab'))
        self.window.ui.tabs['input'].addTab(files, trans('attachments.tab'))
        self.window.ui.tabs['input'].addTab(files_uploaded, trans('attachments_uploaded.tab'))
        self.window.ui.tabs['input'].currentChanged.connect(self.update_min_height)  # update min height on tab change

        self.window.ui.tabs['input'].setTabIcon(0, QIcon(":/icons/input.svg"))
        self.window.ui.tabs['input'].setTabIcon(1, QIcon(":/icons/attachment.svg"))
        self.window.ui.tabs['input'].setTabIcon(2, QIcon(":/icons/upload.svg"))

        # layout
        layout = QVBoxLayout()
        layout.addLayout(self.setup_header())
        layout.addWidget(self.window.ui.tabs['input'])
        layout.addLayout(self.setup_bottom())

        widget = QWidget()
        widget.setLayout(layout)

        return widget

    def setup_input(self) -> QWidget:
        """
        Setup input tab

        :return: QWidget
        """
        # input textarea
        self.window.ui.nodes['input'] = ChatInput(self.window)
        self.window.ui.nodes['input'].setMinimumHeight(self.min_height_input)

        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.nodes['input'])
        layout.setContentsMargins(0, 0, 0, 0)

        widget = QWidget()
        widget.setLayout(layout)
        return widget

    def setup_attachments(self) -> QWidget:
        """
        Setup attachments

        :return: QWidget
        """
        layout = QVBoxLayout()
        layout.addLayout(self.attachments.setup())
        layout.setContentsMargins(0, 0, 0, 0)

        widget = QWidget()
        widget.setLayout(layout)
        widget.setMinimumHeight(self.min_height_files_tab)
        return widget

    def setup_attachments_uploaded(self) -> QWidget:
        """
        Setup attachments uploaded

        :return: QWidget
        """
        layout = QVBoxLayout()
        layout.addLayout(self.attachments_uploaded.setup())
        layout.setContentsMargins(0, 0, 0, 0)

        widget = QWidget()
        widget.setLayout(layout)
        widget.setMinimumHeight(self.min_height_files_tab)
        return widget

    def setup_header(self) -> QHBoxLayout:
        """
        Setup input header

        :return: QHBoxLayout
        """
        # header (input label + input counter)
        self.window.ui.nodes['input.label'] = HelpLabel(trans("input.label"))
        self.window.ui.nodes['input.counter'] = QLabel("")
        self.window.ui.nodes['input.counter'].setToolTip(trans('tip.tokens.input'))

        # plugin audio input addon
        self.window.ui.plugin_addon['audio.input'] = AudioInput(self.window)
        self.window.ui.plugin_addon['audio.input.btn'] = AudioInputButton(self.window)

        header = QHBoxLayout()
        header.addWidget(self.window.ui.nodes['input.label'])
        header.addWidget(self.window.ui.plugin_addon['audio.input'])
        header.addWidget(self.window.ui.plugin_addon['audio.input.btn'])
        header.addWidget(self.window.ui.nodes['input.counter'], alignment=Qt.AlignRight)

        return header

    def setup_bottom(self) -> QHBoxLayout:
        """
        Setup input bottom

        :return: QHBoxLayout
        """
        layout = QHBoxLayout()
        layout.addLayout(self.status.setup())
        layout.addLayout(self.setup_buttons())

        return layout

    def setup_buttons(self) -> QHBoxLayout:
        """
        Setup input buttons

        :return: QHBoxLayout
        """
        # send with: enter
        self.window.ui.nodes['input.send_enter'] = QRadioButton(trans("input.radio.enter"))
        self.window.ui.nodes['input.send_enter'].clicked.connect(
            lambda: self.window.controller.chat.common.toggle_send_shift(
                1))

        # send with: shift + enter
        self.window.ui.nodes['input.send_shift_enter'] = QRadioButton(trans("input.radio.enter_shift"))
        self.window.ui.nodes['input.send_shift_enter'].clicked.connect(
            lambda: self.window.controller.chat.common.toggle_send_shift(
                2))

        # send with: none
        self.window.ui.nodes['input.send_none'] = QRadioButton(trans("input.radio.none"))
        self.window.ui.nodes['input.send_none'].clicked.connect(
            lambda: self.window.controller.chat.common.toggle_send_shift(
                0))

        # send clear
        self.window.ui.nodes['input.send_clear'] = QCheckBox(trans('input.send_clear'))
        self.window.ui.nodes['input.send_clear'].stateChanged.connect(
            lambda: self.window.controller.chat.common.toggle_send_clear(
                self.window.ui.nodes['input.send_clear'].isChecked()))

        # stream
        self.window.ui.nodes['input.stream'] = QCheckBox(trans('input.stream'))
        self.window.ui.nodes['input.stream'].stateChanged.connect(
            lambda: self.window.controller.chat.common.toggle_stream(self.window.ui.nodes['input.stream'].isChecked()))

        # send button
        self.window.ui.nodes['input.send_btn'] = QPushButton(trans("input.btn.send"))
        self.window.ui.nodes['input.send_btn'].clicked.connect(
            lambda: self.window.controller.chat.input.send_input())

        # stop button
        self.window.ui.nodes['input.stop_btn'] = QPushButton(trans("input.btn.stop"))
        self.window.ui.nodes['input.stop_btn'].setVisible(False)
        self.window.ui.nodes['input.stop_btn'].clicked.connect(
            lambda: self.window.controller.chat.common.stop())

        # layout
        self.window.ui.nodes['ui.input.buttons'] = QHBoxLayout()
        self.window.ui.nodes['ui.input.buttons'].addWidget(self.window.ui.nodes['input.stream'])
        self.window.ui.nodes['ui.input.buttons'].addWidget(self.window.ui.nodes['input.send_clear'])
        self.window.ui.nodes['ui.input.buttons'].addWidget(self.window.ui.nodes['input.send_enter'])
        self.window.ui.nodes['ui.input.buttons'].addWidget(self.window.ui.nodes['input.send_shift_enter'])
        self.window.ui.nodes['ui.input.buttons'].addWidget(self.window.ui.nodes['input.send_none'])
        self.window.ui.nodes['ui.input.buttons'].addWidget(self.window.ui.nodes['input.send_btn'])
        self.window.ui.nodes['ui.input.buttons'].addWidget(self.window.ui.nodes['input.stop_btn'])
        self.window.ui.nodes['ui.input.buttons'].setAlignment(Qt.AlignRight)

        return self.window.ui.nodes['ui.input.buttons']

    def update_min_height(self):
        """
        Update the minimum height of the input tab
        """
        idx = self.window.ui.tabs['input'].currentIndex()
        if idx == 0:
            self.window.ui.nodes['input'].setMinimumHeight(self.min_height_input)
            self.window.ui.tabs['input'].setMinimumHeight(self.min_height_input_tab)
        else:
            self.window.ui.nodes['input'].setMinimumHeight(self.min_height_files_tab)
            self.window.ui.tabs['input'].setMinimumHeight(self.min_height_files_tab + 60)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.20 19:00:00                  #
# ================================================== #

import re
from bs4 import BeautifulSoup


class WebSearch:
    def __init__(self, plugin=None):
        """
        Web search

        :param plugin: plugin
        """
        self.plugin = plugin
        self.signals = None

    def search(
            self,
            query: str,
            num: int,
            offset: int = 1
    ) -> list:
        """
        Execute search query and return list of urls

        :param query: query string
        :param num: number of results
        :param offset: offset
        :return: list of URLs
        """
        return self.plugin.get_provider().search(query, num, offset)

    def get_urls(self, query: str) -> list:
        """
        Search the web and returns URLs

        :param query: query string
        :return: list of founded URLs
        """
        urls = []
        try:
            urls = self.search(
                query,
                int(self.plugin.get_option_value("num_pages")),
            )
        except Exception as e:
            self.error(e)
            self.debug(
                "Plugin: web search: error: {}".format(e)
            )
            self.log("Error in web search: " + str(e))
        return urls

    def query_url(self, url: str) -> str:
        """
        Query a URL and return the text content

        :param url: URL to query
        :return: text content
        """
        self.debug(
            "Plugin: cmd_web:query_url: crawling URL: {}".format(url)
        )
        text = ''
        html = ''
        try:
            data = self.plugin.get_url(url)
            try:
                html = data.decode("utf-8")  # try to decode
            except Exception as e:
                pass

            if html:
                soup = BeautifulSoup(html, "html.parser")
                for element in soup.find_all('html'):
                    text += element.text
                text = text.replace("\n", " ").replace("\t", " ")
                text = re.sub(r'\s+', ' ', text)
                self.debug(
                    "Plugin: cmd_web:query_url: received text: {}".format(text)
                )
                return text
        except Exception as e:
            self.error(e)
            self.debug(
                "Plugin: cmd_web:query_url: error querying: {}".format(url)
            )
            self.log("Error in query_web: " + str(e))

    def to_chunks(
            self,
            text: str,
            chunk_size: int
    ) -> list:
        """
        Split text into chunks

        :param text: text to split
        :param chunk_size: chunk size
        :return: list of chunks
        """
        if text is None or text == "":
            return []
        return [text[i:i + chunk_size] for i in range(0, len(text), chunk_size)]

    def get_summary(
            self,
            chunks: list,
            query: str,
            summarize_prompt: str = None
    ) -> str:
        """
        Get summarized text from chunks

        :param chunks: chunks of text
        :param query: query string
        :param summarize_prompt: custom summarize prompt
        :return: summarized text
        """
        summary = ""
        sys_prompt = "Summarize text in English in a maximum of 3 paragraphs, trying to find the most important " \
                     "content that can help answer the following question: {query}".format(query=query)

        # get custom prompt if set
        custom_summarize = self.plugin.get_option_value('prompt_summarize')
        if custom_summarize is not None and custom_summarize != "":
            try:
                sys_prompt = str(custom_summarize.format(query=query))
            except Exception as e:
                self.error(e)
                self.debug(
                    "Plugin: cmd_web:get_summary: prompt error: {}".format(e)
                )
        max_tokens = int(self.plugin.get_option_value("summary_max_tokens"))

        # custom summarize prompt
        if summarize_prompt is not None and summarize_prompt != "":
            sys_prompt = summarize_prompt
        else:
            if query is None or query == "":
                sys_prompt = str(self.plugin.get_option_value(
                    'prompt_summarize_url'
                ).format(query=query))

        # get model
        model = self.plugin.window.core.models.from_defaults()
        tmp_model = self.plugin.get_option_value("summary_model")
        if self.plugin.window.core.models.has(tmp_model):
            model = self.plugin.window.core.models.get(tmp_model)

        # summarize per chunk
        for chunk in chunks:
            # print("Chunk: " + chunk)
            self.debug(
                "Plugin: cmd_web:get_summary (chunk, max_tokens): {}, {}".format(chunk, max_tokens)
            )
            try:
                response = self.plugin.window.core.bridge.quick_call(
                    prompt=chunk,
                    system_prompt=sys_prompt,
                    max_tokens=max_tokens,
                    model=model,
                )
                if response is not None and response != "":
                    summary += response
            except Exception as e:
                self.error(e)
                self.debug(
                    "Plugin: cmd_web:get_summary: error: {}".format(e)
                )
        return summary

    def make_query(
            self,
            query: str,
            page_no: int = 1,
            summarize_prompt: str = ""
    ) -> (str, int, int, str):
        """
        Get result from search query

        :param query: query to search
        :param page_no: page number
        :param summarize_prompt: custom prompt
        :return: result, total_found, current, url
        """
        self.log("Using web query: " + query)
        urls = self.get_urls(query)

        # get options
        max_per_page = int(self.plugin.get_option_value("max_page_content_length"))
        chunk_size = int(self.plugin.get_option_value("chunk_size"))
        max_result_size = int(self.plugin.get_option_value("max_result_length"))

        total_found = len(urls)
        result = ""
        i = 1
        current = 1
        url = ""
        for url in urls:
            if url is None or url == "":
                continue

            # check if requested page number
            if current != page_no:
                current += 1
                continue

            self.log("Web attempt: " + str(i) + " of " + str(len(urls)))
            self.log("URL: " + url)
            content = self.query_url(url)
            if content is None or content == "":
                i += 1
                continue

            self.log("Content found (chars: {}). Please wait...".format(len(content)))
            if 0 < max_per_page < len(content):
                content = content[:max_per_page]

            chunks = self.to_chunks(content, chunk_size)  # it returns list of chunks
            self.debug(
                "Plugin: cmd_web: URL: {}".format(url)
            )
            result = self.get_summary(
                chunks,
                str(query),
                summarize_prompt,
            )
            # if result then stop
            if result is not None and result != "":
                self.log("Summary generated (chars: {})".format(len(result)))

                # index webpage if auto-index is enabled
                self.index_url(url)
                break
            i += 1

        self.debug(
            "Plugin: cmd_web: summary: {}".format(result)
        )
        if result is not None:
            self.debug(
                "Plugin: cmd_web: summary length: {}".format(len(result))
            )

        if len(result) > max_result_size:
            result = result[:max_result_size]

        self.debug(
            "Plugin: cmd_web: result length: {}".format(len(result))
        )

        return \
            result, \
            total_found, \
            current, \
            url

    def open_url(self, url: str, summarize_prompt: str = "") -> (str, str):
        """
        Get result from specified URL

        :param url: URL to visit
        :param summarize_prompt: custom prompt
        :return: result, url
        """
        self.log("Using URL: " + url)

        # get options
        max_per_page = int(self.plugin.get_option_value("max_page_content_length"))
        chunk_size = int(self.plugin.get_option_value("chunk_size"))
        max_result_size = int(self.plugin.get_option_value("max_result_length"))

        self.log("URL: " + url)
        content = self.query_url(url)
        self.log("Content found (chars: {}). Please wait...".format(len(content)))

        # index webpage if auto-index is enabled
        if content:
            self.index_url(url)

        if 0 < max_per_page < len(content):
            content = content[:max_per_page]
        chunks = self.to_chunks(
            content,
            chunk_size,
        )  # it returns list of chunks

        self.debug("Plugin: cmd_web: URL: {}".format(url))  # log

        result = self.get_summary(chunks, "", summarize_prompt)

        if result is not None and result != "":
            self.log("Summary generated (chars: {})".format(len(result)))

        self.debug("Plugin: cmd_web: summary: {}".format(result))  # log
        if result is not None:
            self.debug(
                "Plugin: cmd_web: summary length: {}".format(len(result))
            )

        if len(result) > max_result_size:
            result = result[:max_result_size]

        self.debug(
            "Plugin: cmd_web: result length: {}".format(len(result))
        )

        return result, url

    def open_url_raw(self, url: str) -> (str, str):
        """
        Get raw content from specified URL

        :param url: URL to visit
        :return: result, url
        """
        self.log("Using URL: " + url)

        # get options
        max_result_size = int(self.plugin.get_option_value("max_page_content_length"))

        self.log("URL: " + url)
        result = self.query_url(url)
        self.log("Content found (chars: {}). Please wait...".format(len(result)))

        # index webpage if auto-index is enabled
        if result:
            self.index_url(url)

        # strip if too long
        if 0 < max_result_size < len(result):
            result = result[:max_result_size]

        self.debug(
            "Plugin: cmd_web: result length: {}".format(len(result))
        )

        return result, url

    def index_url(self, url: str):
        """
        Index URL if auto-index is enabled

        :param url: URL to index
        """
        if not self.plugin.get_option_value("auto_index"):
            return

        self.log("Indexing URL: " + url)
        try:
            num, errors = self.plugin.window.core.idx.index_urls(
                self.plugin.get_option_value("idx"),
                [url],  # list of URLs is expected
            )
            if errors:
                self.error(str(errors))
            elif num > 0:
                self.log("Indexed URL: " + url + " to index: " + self.plugin.get_option_value("idx"))
            else:
                self.error("Failed to index URL: " + url)
        except Exception as e:
            self.error(e)

    def error(self, err: any):
        """
        Log error message

        :param err: exception
        """
        if self.signals is not None:
            self.signals.error.emit(err)

    def status(self, msg: str):
        """
        Send status message

        :param msg: status message
        """
        if self.signals is not None:
            self.signals.status.emit(msg)

    def debug(self, msg: str):
        """
        Log debug message

        :param msg: message to log
        """
        if self.signals is not None:
            self.signals.debug.emit(msg)

    def log(self, msg: str):
        """
        Log message

        :param msg: message to log
        """
        if self.signals is not None:
            self.signals.log.emit(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.14 15:00:00                  #
# ================================================== #

import copy

from PySide6.QtCore import Qt
from PySide6.QtGui import QStandardItemModel
from PySide6.QtWidgets import QPushButton, QHBoxLayout, QLabel, QVBoxLayout, QScrollArea, QWidget, QTabWidget, QFrame

from pygpt_net.item.model import ModelItem
from pygpt_net.ui.widget.dialog.model import ModelDialog
from pygpt_net.ui.widget.element.group import CollapsedGroup
from pygpt_net.ui.widget.element.labels import UrlLabel
from pygpt_net.ui.widget.lists.model_editor import ModelEditorList
from pygpt_net.ui.widget.option.checkbox import OptionCheckbox
from pygpt_net.ui.widget.option.combo import OptionCombo
from pygpt_net.ui.widget.option.dictionary import OptionDict
from pygpt_net.ui.widget.option.input import OptionInput, PasswordInput
from pygpt_net.ui.widget.option.slider import OptionSlider
from pygpt_net.ui.widget.option.textarea import OptionTextarea
from pygpt_net.utils import trans


class Models:
    def __init__(self, window=None):
        """
        Models editor dialog

        :param window: Window instance
        """
        self.window = window
        self.dialog_id = "models.editor"

    def setup(self, idx=None):
        """
        Setup editor dialog

        :param idx: current model tab index
        """
        self.window.ui.nodes['models.editor.btn.new'] = \
            QPushButton(trans("dialog.models.editor.btn.new"))
        self.window.ui.nodes['models.editor.btn.defaults.user'] = \
            QPushButton(trans("dialog.models.editor.btn.defaults.user"))
        self.window.ui.nodes['models.editor.btn.defaults.app'] = \
            QPushButton(trans("dialog.models.editor.btn.defaults.app"))
        self.window.ui.nodes['models.editor.btn.save'] = \
            QPushButton(trans("dialog.models.editor.btn.save"))

        self.window.ui.nodes['models.editor.btn.new'].clicked.connect(
            lambda: self.window.controller.model.editor.new())
        self.window.ui.nodes['models.editor.btn.defaults.user'].clicked.connect(
            lambda: self.window.controller.model.editor.load_defaults_user())
        self.window.ui.nodes['models.editor.btn.defaults.app'].clicked.connect(
            lambda: self.window.controller.model.editor.load_defaults_app())
        self.window.ui.nodes['models.editor.btn.save'].clicked.connect(
            lambda: self.window.controller.model.editor.save())

        # set enter key to save button
        self.window.ui.nodes['models.editor.btn.new'].setAutoDefault(False)
        self.window.ui.nodes['models.editor.btn.defaults.user'].setAutoDefault(False)
        self.window.ui.nodes['models.editor.btn.defaults.app'].setAutoDefault(False)
        self.window.ui.nodes['models.editor.btn.save'].setAutoDefault(True)

        # footer buttons
        footer = QHBoxLayout()
        footer.addWidget(self.window.ui.nodes['models.editor.btn.new'])
        footer.addWidget(self.window.ui.nodes['models.editor.btn.defaults.user'])
        footer.addWidget(self.window.ui.nodes['models.editor.btn.defaults.app'])
        footer.addWidget(self.window.ui.nodes['models.editor.btn.save'])

        # editor tabs
        self.window.ui.tabs['models.editor'] = QTabWidget()

        # build settings tabs
        parent_id = "model"

        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        content = QVBoxLayout()

        # create config options entry if not exists
        if parent_id not in self.window.ui.config:
            self.window.ui.config[parent_id] = {}

        # get options
        options = copy.deepcopy(self.window.controller.model.editor.get_options())
        widgets = self.build_widgets(options)
        advanced_keys = []
        for key in options:
            if 'advanced' in options[key] and options[key]['advanced']:
                advanced_keys.append(key)

        # apply settings widgets
        for key in widgets:
            self.window.ui.config[parent_id][key] = widgets[key]

        for key in widgets:
            if key in advanced_keys:  # hide advanced options
                continue
            content.addLayout(self.add_option(widgets[key], options[key]))  # add to scroll

        # append advanced options at the end
        if len(advanced_keys) > 0:
            group_id = 'models.editor.advanced'
            self.window.ui.groups[group_id] = CollapsedGroup(self.window, group_id, None, False, None)
            self.window.ui.groups[group_id].box.setText(trans('settings.advanced.collapse'))
            for key in widgets:
                if key not in advanced_keys:  # ignore non-advanced options
                    continue

                option = self.add_option(widgets[key], options[key])  # build option
                self.window.ui.groups[group_id].add_layout(option)  # add option to group

            # add advanced options group to scroll
            content.addWidget(self.window.ui.groups[group_id])

        content.addStretch()

        # scroll widget
        scroll_widget = QWidget()
        scroll_widget.setLayout(content)
        scroll.setWidget(scroll_widget)

        area = QVBoxLayout()
        area.addWidget(scroll)

        area_widget = QWidget()
        area_widget.setLayout(area)

        data = {}
        for model_id in self.window.core.models.items.keys():
            model = self.window.core.models.items[model_id]
            data[model_id] = model

        # models list
        id = 'models.list'
        self.window.ui.nodes[id] = ModelEditorList(self.window, id)
        self.window.ui.models[id] = self.create_model(self.window)
        self.window.ui.nodes[id].setModel(self.window.ui.models[id])

        # update models list
        self.update_list(id, data)

        # set max width to list
        self.window.ui.nodes[id].setMaximumWidth(250)

        main_layout = QHBoxLayout()
        main_layout.addWidget(self.window.ui.nodes[id])
        main_layout.addWidget(area_widget)

        layout = QVBoxLayout()
        layout.addLayout(main_layout)  # list + model tabs
        layout.addLayout(footer)  # bottom buttons (save, defaults)

        self.window.ui.dialog[self.dialog_id] = ModelDialog(self.window, self.dialog_id)
        self.window.ui.dialog[self.dialog_id].setLayout(layout)
        self.window.ui.dialog[self.dialog_id].setWindowTitle(trans('dialog.models.editor'))

        # restore current opened tab if idx is set
        if idx is not None:
            try:
                self.window.controller.model.editor.set_by_tab(idx)
            except Exception as e:
                print('Failed restore models editor tab: {}'.format(idx))
        else:
            if self.window.controller.model.editor.current is None:
                self.window.controller.model.editor.set_by_tab(0)

    def build_widgets(self, options: dict) -> dict:
        """
        Build settings options widgets

        :param options: model options
        :return: dict of widgets
        """
        parent = "model"
        widgets = {}

        for key in options:
            option = options[key]
            option['id'] = key
            # create widget by option type
            if option['type'] == 'text' or option['type'] == 'int' or option['type'] == 'float':
                if 'slider' in option and option['slider'] \
                        and (option['type'] == 'int' or option['type'] == 'float'):
                    widgets[key] = OptionSlider(self.window, parent, key, option)  # slider + text input
                else:
                    if 'secret' in option and option['secret']:
                        widgets[key] = PasswordInput(self.window, parent, key, option)  # password input
                    else:
                        widgets[key] = OptionInput(self.window, parent, key, option)  # text input
            elif option['type'] == 'textarea':
                widgets[key] = OptionTextarea(self.window, parent, key, option)  # textarea
            elif option['type'] == 'bool':
                widgets[key] = OptionCheckbox(self.window, parent, key, option)  # checkbox
            elif option['type'] == 'dict':
                self.window.controller.config.placeholder.apply(option)
                widgets[key] = OptionDict(self.window, parent, key, option)  # dictionary
                widgets[key].setMinimumHeight(200)
            elif option['type'] == 'combo':
                self.window.controller.config.placeholder.apply(option)
                widgets[key] = OptionCombo(self.window, parent, key, option)  # combobox

        return widgets

    def add_line(self) -> QFrame:
        """
        Make separator line

        :return: separator line
        """
        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        return line

    def add_urls(self, urls: dict) -> QWidget:
        """
        Add clickable urls to list

        :param urls: urls dict
        :return: QWidget
        """
        layout = QVBoxLayout()
        for name in urls:
            url = urls[name]
            label = UrlLabel(name, url)
            layout.addWidget(label)

        widget = QWidget()
        widget.setLayout(layout)
        return widget

    def add_option(self, widget: QWidget, option: dict) -> QVBoxLayout:
        """
        Append option widget to layout

        :param widget: widget instance
        :param option: option dict
        :return: QVBoxLayout
        """
        one_column_types = ['textarea', 'dict', 'bool']
        key = option['id']
        label = trans(option['label'])
        label_key = 'model.' + key + '.label'
        desc = None
        desc_key = None

        if option['type'] != 'bool':
            self.window.ui.nodes[label_key] = QLabel(label)
            self.window.ui.nodes[label_key].setStyleSheet("font-weight: bold;")

        if "description" in option:
            desc = trans(option['description'])
            desc_key = 'model.' + key + '.desc'

        # 2-columns layout
        if option['type'] not in one_column_types:
            cols = QHBoxLayout()
            cols.addWidget(self.window.ui.nodes[label_key])  # disable label in bool type
            cols.addWidget(widget)

            cols_widget = QWidget()
            cols_widget.setLayout(cols)
            cols_widget.setMaximumHeight(90)

            layout = QVBoxLayout()
            layout.addWidget(cols_widget)
        else:
            # 1-column layout: textarea and dict fields
            layout = QVBoxLayout()
            if option['type'] != 'bool':
                layout.addWidget(self.window.ui.nodes[label_key])
            else:
                widget.box.setText(label)  # set checkbox label
            layout.addWidget(widget)

        if desc:
            self.window.ui.nodes[desc_key] = QLabel(desc)
            self.window.ui.nodes[desc_key].setWordWrap(True)
            self.window.ui.nodes[desc_key].setMaximumHeight(40)
            self.window.ui.nodes[desc_key].setStyleSheet("font-size: 10px;")
            layout.addWidget(self.window.ui.nodes[desc_key])

        line = self.add_line()  # TODO: change name to separator
        layout.addWidget(line)

        return layout

    def add_raw_option(self, option: dict) -> QHBoxLayout:
        """
        Add raw option row

        :param option: Option
        :return: QHBoxLayout
        """
        layout = QHBoxLayout()
        layout.addWidget(option)
        return layout

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model
        :param parent: parent widget
        :return: QStandardItemModel
        """
        return QStandardItemModel(0, 1, parent)

    def update_list(self, id: str, data: dict):
        """
        Update list

        :param id: ID of the list
        :param data: Data to update
        """
        self.window.ui.models[id].removeRows(0, self.window.ui.models[id].rowCount())
        i = 0
        for n in data:
            self.window.ui.models[id].insertRow(i)
            name = data[n].name
            self.window.ui.models[id].setData(self.window.ui.models[id].index(i, 0), name)
            i += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.21 14:00:00                  #
# ================================================== #

import re
import html
from datetime import datetime
from PySide6.QtGui import QTextCursor, QTextBlockFormat, QTextCharFormat

from pygpt_net.core.render.base import BaseRenderer
from pygpt_net.item.ctx import CtxItem
from pygpt_net.ui.widget.textarea.input import ChatInput
from pygpt_net.ui.widget.textarea.output import ChatOutput
from pygpt_net.utils import trans
from .parser import Parser


class Renderer(BaseRenderer):
    def __init__(self, window=None):
        super(Renderer, self).__init__(window)
        """
        Markdown renderer

        :param window: Window instance
        """
        self.window = window
        self.parser = Parser(window)
        self.images_appended = []
        self.urls_appended = []
        self.buffer = ""
        self.is_cmd = False
        self.img_width = 400

    def begin(self, stream: bool = False):
        """
        Render begin

        :param stream: True if it is a stream
        """
        pass  # do nothing

    def end(self, stream: bool = False):
        """
        Render end

        :param stream: True if it is a stream
        """
        if stream:
            self.reload()  # reload ctx items only if stream

    def end_extra(self, stream: bool = False):
        """
        Render end extra

        :param stream: True if it is a stream
        """
        self.to_end()

    def stream_begin(self):
        """Render stream begin"""
        pass  # do nothing

    def stream_end(self):
        """Render stream end"""
        pass  # do nothing

    def clear_output(self):
        """Clear output"""
        self.reset()
        self.get_output_node().clear()

    def clear_input(self):
        """Clear input"""
        self.get_input_node().clear()

    def reset(self):
        """Reset"""
        self.images_appended = []
        self.urls_appended = []

    def reload(self):
        """Reload output, called externally only on theme change to redraw content"""
        self.window.controller.ctx.refresh_output()  # if clear all and appends all items again

    def append_context(self, items: list, clear: bool = True):
        """
        Append all context to output

        :param items: Context items
        :param clear: True if clear all output before append
        """
        if clear:
            self.clear_output()
        i = 0
        for item in items:
            item.idx = i
            if i == 0:
                item.first = True
            self.append_context_item(item)
            i += 1

    def append_input(self, item: CtxItem):
        """
        Append text input to output

        :param item: context item
        """
        if item.input is None or item.input == "":
            return
        if self.is_timestamp_enabled() \
                and item.input_timestamp is not None:
            name = ""
            if item.input_name is not None \
                    and item.input_name != "":
                name = item.input_name + " "
            text = '{} > {}'.format(name, item.input)
        else:
            text = "> {}".format(item.input)

        # check if it is a command response
        is_cmd = False
        if item.input.strip().startswith("[") \
                and item.input.strip().endswith("]"):
            is_cmd = True

        # hidden internal call
        if item.internal \
                and not is_cmd \
                and not item.first \
                and not item.input.strip().startswith("user: "):
            self.append_raw('>>>', "msg-user", item)
            return
        else:
            # don't show user prefix if provided in internal call goal update
            if item.internal and item.input.startswith("user: "):
                text = re.sub(r'^user: ', '> ', item.input)

        self.append_raw(text.strip(), "msg-user", item)

    def append_output(self, item: CtxItem):
        """
        Append text output to output

        :param item: context item
        """
        if item.output is None or item.output == "":
            return
        if self.is_timestamp_enabled() \
                and item.output_timestamp is not None:
            name = ""
            if item.output_name is not None \
                    and item.output_name != "":
                name = item.output_name + " "
            text = '{} {}'.format(name, item.output)
        else:
            text = "{}".format(item.output)
        self.append_raw(text.strip(), "msg-bot", item)

    def append_extra(self, item: CtxItem):
        """
        Append extra data (images, files, etc.) to output

        :param item: context item
        """
        appended = []

        # images
        if len(item.images) > 0:
            for image in item.images:
                # don't append if it is a external url
                if image.startswith("http"):
                    continue
                if image in appended or image in self.images_appended:
                    continue
                try:
                    appended.append(image)
                    self.get_output_node().append(self.get_image_html(image))
                    self.images_appended.append(image)
                except Exception as e:
                    pass

        # files and attachments, TODO check attachments
        if len(item.files) > 0:
            for file in item.files:
                if file in appended:
                    continue
                try:
                    appended.append(file)
                    self.get_output_node().append(self.get_file_html(file))
                except Exception as e:
                    pass

        # urls
        if len(item.urls) > 0:
            for url in item.urls:
                if url in appended or url in self.urls_appended:
                    continue
                try:
                    appended.append(url)
                    self.get_output_node().append(self.get_url_html(url))
                    self.urls_appended.append(url)
                except Exception as e:
                    pass

        # jump to end
        if len(appended) > 0:
            self.to_end()

    def append_chunk(self, item: CtxItem, text_chunk: str, begin: bool = False):
        """
        Append output chunk to output

        :param item: context item
        :param text_chunk: text chunk
        :param begin: if it is the beginning of the text
        """
        if text_chunk is None or text_chunk == "":
            return

        raw_chunk = str(text_chunk)

        if begin:
            self.buffer = ""  # reset buffer
            self.is_cmd = False  # reset command flag

            if self.is_timestamp_enabled() \
                    and item.output_timestamp is not None:
                name = ""
                if item.output_name is not None \
                        and item.output_name != "":
                    name = item.output_name + " "
                ts = datetime.fromtimestamp(item.output_timestamp)
                hour = ts.strftime("%H:%M:%S")
                text_chunk = "{}{}: ".format(name, hour) + text_chunk

            self.append_block()
            self.append_chunk_start()

        self.buffer += raw_chunk
        self.append(self.format_chunk(text_chunk), "")

    def append_block(self):
        """Append block to output"""
        cursor = self.get_output_node().textCursor()
        cursor.movePosition(QTextCursor.End)
        default_format = QTextCharFormat()  # reset format
        cursor.setCharFormat(default_format)
        block_format = QTextBlockFormat()
        block_format.setIndent(0)
        cursor.insertBlock(block_format)
        self.get_output_node().setTextCursor(cursor)

    def to_end(self):
        """Move cursor to end of output"""
        cursor = self.get_output_node().textCursor()
        cursor.movePosition(QTextCursor.End)
        self.get_output_node().setTextCursor(cursor)

    def append_raw(self, text: str, type: str = "msg-bot", item: CtxItem = None):
        """
        Append and format raw text to output

        :param text: text to append
        :param type: type of message
        :param item: CtxItem instance
        """
        if type != "msg-user":  # markdown for bot messages
            text = self.pre_format_text(text)
            text = self.append_timestamp(text, item)
            text = self.parser.parse(text)
        else:
            content = self.append_timestamp(self.format_user_text(text), item, type=type)
            text = "<div><p>" + content + "</p></div>"

        text = self.post_format_text(text)
        text = '<div class="{}">'.format(type) + text.strip() + "</div>"

        self.get_output_node().append(text)
        self.to_end()

    def append_chunk_start(self):
        """
        Append start of chunk to output
        """
        cursor = self.get_output_node().textCursor()
        cursor.movePosition(QTextCursor.End)
        self.get_output_node().setTextCursor(cursor)

    def append_context_item(self, item: CtxItem):
        """
        Append context item to output

        :param item: context item
        """
        self.append_input(item)
        self.append_output(item)
        self.append_extra(item)

    def get_image_html(self, url: str) -> str:
        """
        Get image HTML

        :param url: URL to image
        :return: HTML code
        """
        url, path = self.window.core.filesystem.extract_local_url(url)
        return """<a href="{url}"><img src="{path}" width="{img_width}" class="image"></a>
        <p><b>{prefix}:</b> <a href="{url}">{path}</a></p>""".\
            format(prefix=trans('chat.prefix.img'),
                   url=url,
                   path=path,
                   img_width=self.img_width)

    def get_url_html(self, url: str) -> str:
        """
        Get URL HTML

        :param url: external URL
        :return: HTML code
        """
        return """<br/><b>{prefix}:</b> <a href="{url}">{url}</a><br/>""".\
            format(prefix=trans('chat.prefix.url'),
                   url=url)

    def get_file_html(self, url: str) -> str:
        """
        Get file HTML

        :param url: URL to file
        :return: HTML code
        """
        url, path = self.window.core.filesystem.extract_local_url(url)
        return """<div><b>{prefix}:</b> <a href="{url}">{path}</a></div>""".\
            format(prefix=trans('chat.prefix.file'),
                   url=url,
                   path=path)

    def append(self, text: str, end: str = "\n"):
        """
        Append text to output

        :param text: text to append
        :param end: end of the line character
        """
        cur = self.get_output_node().textCursor()  # Move cursor to end of text
        cur.movePosition(QTextCursor.End)
        s = str(text) + end
        while s:
            head, sep, s = s.partition("\n")  # Split line at LF
            cur.insertText(head)  # Insert text at cursor
            if sep:  # New line if LF
                cur.insertHtml("<br>")
        self.get_output_node().setTextCursor(cur)  # Update visible cursor

    def append_timestamp(self, text: str, item: CtxItem, type: str = None) -> str:
        """
        Append timestamp to text

        :param text: Input text
        :param item: Context item
        :param type: Type of message
        :return: Text with timestamp (if enabled)
        """
        if item is not None \
                and self.is_timestamp_enabled() \
                and item.input_timestamp is not None:
            if type == "msg-user":
                timestamp = item.input_timestamp
            else:
                timestamp = item.output_timestamp
            if timestamp is not None:
                ts = datetime.fromtimestamp(timestamp)
                hour = ts.strftime("%H:%M:%S")
                text = '<span class="ts">{}:</span> {}'.format(hour, text)
        return text

    def replace_code_tags(self, text: str) -> str:
        """
        Replace cmd code tags

        :param text:
        :return: replaced text
        """
        pattern = r"~###~(.*?)~###~"
        replacement = r'<p class="cmd">\1</p>'
        return re.sub(pattern, replacement, text)

    def pre_format_text(self, text: str) -> str:
        """
        Pre-format text

        :param text: text to format
        :return: formatted text
        """
        text = text.strip()
        text = text.replace("#~###~", "~###~")  # fix for #~###~ in text (previous versions)
        text = text.replace("# ~###~", "~###~")  # fix for # ~###~ in text (previous versions)

        # replace cmd tags
        text = self.replace_code_tags(text)

        # replace %workdir% with current workdir
        local_prefix = self.window.core.filesystem.get_workdir_prefix()
        safe_local_prefix = local_prefix.replace('\\', '\\\\').replace('\\.', '\\\\.')  # windows fix
        replacement = f'({safe_local_prefix}\\1)'
        try:
            text = re.sub(r'\(%workdir%([^)]+)\)', replacement, text)
        except Exception as e:
            pass
        return text

    def post_format_text(self, text: str) -> str:
        """
        Post-format text

        :param text: text to format
        :return: formatted text
        """
        return text.strip()

    def format_user_text(self, text: str) -> str:
        """
        Post-format user text

        :param text: text to format
        :return: formatted text
        """
        text = html.escape(text).replace("\n", "<br>")

        # append cmd tags if response from command detected
        if text.strip().startswith("&gt; [") and text.strip().endswith("]"):
            text = '<div class="cmd">&gt; {}</div><br/>'.format(text)
        return text

    def format_chunk(self, text: str) -> str:
        """
        Format chunk

        :param text: text to format
        :return: formatted text
        """
        return text

    def is_timestamp_enabled(self) -> bool:
        """
        Check if timestamp is enabled

        :return: True if timestamp is enabled
        """
        return self.window.core.config.get('output_timestamp')

    def get_output_node(self) -> ChatOutput:
        """
        Get output node

        :return: output node
        """
        return self.window.ui.nodes['output']

    def get_input_node(self) -> ChatInput:
        """
        Get input node

        :return: input node
        """
        return self.window.ui.nodes['input']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2023.12.31 04:00:00                  #
# ================================================== #

class Launcher:
    def __init__(self, window=None):
        """
        Launcher controller

        :param window: Window instance
        """
        self.window = window

    def post_setup(self):
        """Post setup launcher"""
        # show welcome API KEY dialog (disable for langchain mode)
        if self.window.core.config.get('mode') != 'langchain' and \
                (self.window.core.config.get('api_key') is None or self.window.core.config.get('api_key') == ''):
            self.show_api_monit()

        # check for updates
        if self.window.core.config.get('updater.check.launch'):
            self.window.core.updater.check()

    def show_api_monit(self):
        """Show empty API KEY monit"""
        self.window.ui.dialogs.open('info.start')

    def check_updates(self):
        """Check for updates"""
        self.window.core.updater.check(True)

    def toggle_update_check(self, value):
        """Toggle update check on startup"""
        self.window.core.config.set('updater.check.launch', value)
        self.window.core.config.save()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.27 11:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QVBoxLayout, QWidget

from pygpt_net.ui.layout.ctx.search_input import SearchInput
from pygpt_net.ui.layout.ctx.ctx_list import CtxList
from pygpt_net.ui.layout.ctx.video import Video
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.utils import trans


class CtxMain:
    def __init__(self, window=None):
        """
        Context list UI

        :param window: Window instance
        """
        self.window = window
        self.search_input = SearchInput(window)
        self.ctx_list = CtxList(window)
        self.video = Video(window)

    def setup(self) -> QWidget:
        """
        Setup layout

        :return: QWidget
        """
        ctx = self.ctx_list.setup()
        video = self.video.setup()
        search_input = self.search_input.setup()

        self.window.ui.nodes['tip.toolbox.ctx'] = HelpLabel(trans('tip.toolbox.ctx'), self.window)

        layout = QVBoxLayout()
        layout.addWidget(ctx)

        layout.addWidget(self.window.ui.nodes['tip.toolbox.ctx'])
        layout.addWidget(search_input)
        layout.addWidget(video)

        widget = QWidget()
        widget.setLayout(layout)

        return widget




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.12 21:00:00                  #
# ================================================== #


from PySide6.QtCore import Qt
from PySide6.QtWidgets import QPushButton


class ContextMenuButton(QPushButton):
    def __init__(self, title, parent=None, action=None):
        super().__init__(title, parent)
        self.action = action

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton or event.button() == Qt.RightButton:
            self.action(self, event.pos())
        else:
            super().mousePressEvent(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.17 17:00:00                  #
# ================================================== #

from pygpt_net.item.ctx import CtxItem


class BaseRenderer:
    def __init__(self, window=None):
        """
        Base renderer

        :param window: Window instance
        """
        self.window = window

    def begin(self, stream: bool = False):
        """Render begin"""
        pass

    def end(self, stream: bool = False):
        """Render end"""
        pass

    def end_extra(self, stream: bool = False):
        """Render end extra"""
        pass

    def stream_begin(self):
        """Render stream begin"""
        pass  # do nothing

    def stream_end(self):
        """Render stream end"""
        pass  # do nothing    

    def clear_output(self):
        """Clear output"""
        pass

    def clear_input(self):
        """Clear input"""
        pass

    def reset(self):
        """Reset"""
        pass

    def reload(self):
        """Reload output, called externally only on theme change to redraw content"""
        pass

    def append_context(self, items: list, clear: bool = True):
        """
        Append all context to output

        :param items: Context items
        :param clear: True if clear all output before append
        """
        pass

    def append_input(self, item: CtxItem):
        """
        Append text input to output

        :param item: context item
        """
        pass

    def append_output(self, item: CtxItem):
        """
        Append text output to output

        :param item: context item
        """
        pass

    def append_extra(self, item: CtxItem):
        """
        Append extra data (images, files, etc.) to output

        :param item: context item
        """
        pass

    def append_chunk(self, item: CtxItem, text_chunk: str, begin: bool = False):
        """
        Append output chunk to output

        :param item: context item
        :param text_chunk: text chunk
        :param begin: if it is the beginning of the text
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.23 06:00:00                  #
# ================================================== #

import uuid
import time

from sqlalchemy import text

from pygpt_net.item.index import IndexItem
from .utils import unpack_file_item


class Storage:
    def __init__(self, window=None):
        """
        DB storage for index items

        :param window: Window instance
        """
        self.window = window

    def attach(self, window):
        """
        Attach window instance

        :param window: Window instance
        """
        self.window = window

    def get_items(self, store_id: str) -> dict:
        """
        Return idx items by store ID

        :return: dict of indexes and files in index
        """
        stmt = text("""
            SELECT * FROM idx_file WHERE store = :store_id ORDER BY id ASC
        """).bindparams(store_id=store_id)
        db = self.window.core.db.get_db()
        items = {}
        with db.connect() as conn:
            result = conn.execute(stmt)
            for row in result:
                idx, data = unpack_file_item(row._asdict())
                if idx not in items:
                    items[idx] = {}
                file_id = data['name']  # use file name as ID
                items[idx][file_id] = data

        indexes = {}
        for idx, files in items.items():
            index = IndexItem()
            index.id = idx
            index.store = store_id
            index.items = files
            indexes[idx] = index

        return indexes

    def insert_file(self, store_id: str, idx: str, data: dict) -> int:
        """
        Insert file to index

        :param store_id: store ID
        :param idx: index
        :param data: dictionary with file data
        """
        id = None
        db = self.window.core.db.get_db()
        stmt = text("""
            INSERT INTO idx_file
            (
                uuid,
                doc_id,
                created_ts,
                updated_ts,
                name,
                path,
                store,
                idx
            )
            VALUES 
            (
                :uuid,
                :doc_id,
                :created_ts,
                :updated_ts,
                :name,
                :path,
                :store,
                :idx
            )
        """).bindparams(
            uuid=str(uuid.uuid4()),
            doc_id=data['id'],
            created_ts=int(data["indexed_ts"]),
            updated_ts=int(data["indexed_ts"]),
            name=data['name'],
            path=data['path'],
            store=store_id,
            idx=idx,
        )
        with db.begin() as conn:
            result = conn.execute(stmt)
            id = result.lastrowid

        return id

    def insert_ctx_meta(self, store_id: str, idx: str, meta_id: int, doc_id: str) -> int:
        """
        Insert ctx meta to index

        :param store_id: store ID
        :param idx: index
        :param meta_id: meta ID
        :param doc_id: document ID
        """
        id = None
        ts = int(time.time())
        db = self.window.core.db.get_db()
        stmt = text("""
            INSERT INTO idx_ctx
            (
                uuid,
                meta_id,
                doc_id,
                created_ts,
                updated_ts,
                store,
                idx
            )
            VALUES 
            (
                :uuid,
                :meta_id,
                :doc_id,
                :created_ts,
                :updated_ts,
                :store,
                :idx
            )
        """).bindparams(
            uuid=str(uuid.uuid4()),
            meta_id=meta_id,
            doc_id=doc_id,
            created_ts=ts,
            updated_ts=ts,
            store=store_id,
            idx=idx,
        )
        with db.begin() as conn:
            result = conn.execute(stmt)
            id = result.lastrowid

        return id

    def insert_external(self, store_id: str, idx: str, data: dict) -> int:
        """
        Insert external data to index

        :param store_id: store ID
        :param idx: index
        :param data: dictionary with external data
        """
        id = None
        db = self.window.core.db.get_db()
        stmt = text("""
            INSERT INTO idx_external
            (
                uuid,
                doc_id,
                created_ts,
                updated_ts,
                content,
                type,
                store,
                idx
            )
            VALUES 
            (
                :uuid,
                :doc_id,
                :created_ts,
                :updated_ts,
                :content,
                :type,
                :store,
                :idx
            )
        """).bindparams(
            uuid=str(uuid.uuid4()),
            doc_id=data['id'],
            created_ts=int(data["indexed_ts"]),
            updated_ts=int(data["indexed_ts"]),
            content=data['content'],
            type=data['type'],
            store=store_id,
            idx=idx,
        )
        with db.begin() as conn:
            result = conn.execute(stmt)
            id = result.lastrowid

        return id

    def is_meta_indexed(self, store_id: str, idx: str, meta_id: int) -> bool:
        """
        Check if context meta is indexed

        :param: store_id: store ID
        :param: idx: index
        :param: meta_id: context meta ID
        :return: True if indexed
        """
        db = self.window.core.db.get_db()
        stmt = text("""
            SELECT COUNT(*) as count
            FROM idx_ctx
            WHERE store = :store_id
            AND idx = :idx
            AND meta_id = :meta_id
        """).bindparams(
            store_id=store_id,
            idx=idx,
            meta_id=meta_id,
        )
        with db.connect() as conn:
            result = conn.execute(stmt)
            row = result.fetchone()
            data = row._asdict()
            return int(data['count']) > 0

    def is_file_indexed(self, store_id: str, idx: str, file_id: str) -> bool:
        """
        Check if file is indexed

        :param: store_id: store ID
        :param: idx: index
        :param: file_id: file ID
        :return: True if indexed
        """
        db = self.window.core.db.get_db()
        stmt = text("""
            SELECT COUNT(*) as count
            FROM idx_file
            WHERE store = :store_id
            AND idx = :idx
            AND name = :file_id
        """).bindparams(
            store_id=store_id,
            idx=idx,
            file_id=file_id,
        )
        with db.connect() as conn:
            result = conn.execute(stmt)
            row = result.fetchone()
            data = row._asdict()
            return int(data['count']) > 0

    def is_external_indexed(self, store_id: str, idx: str, content: str, type: str) -> bool:
        """
        Check if external is indexed

        :param: store_id: store ID
        :param: idx: index
        :param: content: content
        :param: type: type
        :return: True if indexed
        """
        db = self.window.core.db.get_db()
        stmt = text("""
            SELECT COUNT(*) as count
            FROM idx_external
            WHERE store = :store_id
            AND idx = :idx
            AND content = :content
            AND type = :type
        """).bindparams(
            store_id=store_id,
            idx=idx,
            content=content,
            type=type,
        )
        with db.connect() as conn:
            result = conn.execute(stmt)
            row = result.fetchone()
            data = row._asdict()
            return int(data['count']) > 0

    def get_meta_doc_id(self, store_id: str, idx: str, meta_id: int) -> str:
        """
        Get indexed document id by meta id

        :param store_id: store id
        :param idx: index name
        :param meta_id: meta id
        :return: document id
        """
        db = self.window.core.db.get_db()
        stmt = text("""
            SELECT doc_id
            FROM idx_ctx
            WHERE store = :store_id
            AND idx = :idx
            AND meta_id = :meta_id
        """).bindparams(
            store_id=store_id,
            idx=idx,
            meta_id=meta_id,
        )
        with db.connect() as conn:
            result = conn.execute(stmt)
            row = result.fetchone()
            data = row._asdict()
            return data['doc_id']

    def get_file_doc_id(self, store_id: str, idx: str, file_id: str) -> str:
        """
        Get indexed document id by file id

        :param store_id: store id
        :param idx: index name
        :param file_id
        :return: document id
        """
        db = self.window.core.db.get_db()
        stmt = text("""
            SELECT doc_id
            FROM idx_file
            WHERE store = :store_id
            AND idx = :idx
            AND name = :file_id
        """).bindparams(
            store_id=store_id,
            idx=idx,
            file_id=file_id,
        )
        with db.connect() as conn:
            result = conn.execute(stmt)
            row = result.fetchone()
            data = row._asdict()
            return data['doc_id']

    def get_external_doc_id(self, store_id: str, idx: str, content: str, type: str) -> str:
        """
        Get indexed document id by external

        :param store_id: store id
        :param idx: index name
        :param content: content
        :param type: type
        :return: document id
        """
        db = self.window.core.db.get_db()
        stmt = text("""
            SELECT doc_id
            FROM idx_external
            WHERE store = :store_id
            AND idx = :idx
            AND content = :content
            AND type = :type
        """).bindparams(
            store_id=store_id,
            idx=idx,
            content=content,
            type=type,
        )
        with db.connect() as conn:
            result = conn.execute(stmt)
            row = result.fetchone()
            data = row._asdict()
            return data['doc_id']

    def update_file(self, id: int, doc_id: str, ts: int) -> bool:
        """
        Update timestamp of file in index

        :param id: db record ID
        :param doc_id: document ID
        :param ts: timestamp
        """
        db = self.window.core.db.get_db()
        stmt = text("""
            UPDATE idx_file
            SET 
            updated_ts = :updated_ts,
            doc_id = :doc_id
            WHERE id = :id
        """).bindparams(
            id=id,
            doc_id=doc_id,
            updated_ts=ts,
        )
        with db.begin() as conn:
            conn.execute(stmt)
        return True

    def update_ctx_meta(self, meta_id: int, doc_id: str) -> bool:
        """
        Update timestamp of ctx meta in index

        :param meta_id: ctx meta ID
        :param doc_id: document ID
        """
        db = self.window.core.db.get_db()
        stmt = text("""
            UPDATE 
                idx_ctx
            SET 
                updated_ts = :updated_ts,
                doc_id = :doc_id
            WHERE meta_id = :id
        """).bindparams(
            id=meta_id,
            doc_id=doc_id,
            updated_ts=int(time.time()),
        )
        with db.begin() as conn:
            conn.execute(stmt)
        return True

    def update_external(self, content: str, type: str, doc_id: str, ts: int) -> bool:
        """
        Update timestamp of external data in index

        :param content: content
        :param type: type
        :param doc_id: document ID
        :param ts: timestamp
        """
        db = self.window.core.db.get_db()
        stmt = text("""
            UPDATE idx_external
            SET 
            updated_ts = :updated_ts,
            doc_id = :doc_id
            WHERE 
            content = :content AND type = :type
        """).bindparams(
            content=content,
            type=type,
            doc_id=doc_id,
            updated_ts=ts,
        )
        with db.begin() as conn:
            conn.execute(stmt)
        return True

    def remove_file(self, store_id: str, idx: str, doc_id: str):
        """
        Remove file from index

        :param store_id: store ID
        :param idx: index
        :param doc_id: document ID
        """
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(
                text("DELETE FROM idx_file WHERE store = :store_id AND idx = :idx AND doc_id = :doc_id").bindparams(
                    store_id=store_id,
                    idx=idx,
                    doc_id=doc_id,
                ))

    def remove_ctx_meta(self, store_id: str, idx: str, meta_id: str):
        """
        Remove file from index

        :param store_id: store ID
        :param idx: index
        :param meta_id: ctx meta ID
        """
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(
                text("DELETE FROM idx_ctx WHERE store = :store_id AND idx = :idx AND meta_id = :meta_id").bindparams(
                    store_id=store_id,
                    idx=idx,
                    meta_id=meta_id,
                ))

    def remove_external(self, store_id: str, idx: str, doc_id: str):
        """
        Remove file from index

        :param store_id: store ID
        :param idx: index
        :param doc_id: document ID
        """
        db = self.window.core.db.get_db()
        with db.begin() as conn:
            conn.execute(
                text("DELETE FROM idx_external WHERE store = :store_id AND idx = :idx AND doc_id = :doc_id").bindparams(
                    store_id=store_id,
                    idx=idx,
                    doc_id=doc_id,
                ))

    def truncate_all(self, store_id: str = None, idx: str = None) -> bool:
        """
        Truncate all idx tables in database (all stores)

        :param store_id: store ID
        :param idx: index ID
        :return: True if truncated
        """
        self.truncate_files(store_id, idx)
        self.truncate_ctx(store_id, idx)
        self.truncate_external(store_id, idx)
        return True

    def truncate_files(self, store_id: str = None, idx: str = None) -> bool:
        """
        Truncate files table in database

        :param store_id: store ID
        :param idx: index ID
        :return: True if truncated
        """
        return self.truncate_by_db_table("idx_file", store_id, idx)

    def truncate_ctx(self, store_id: str = None, idx: str = None) -> bool:
        """
        Truncate context table in database

        :param store_id: store ID
        :param idx: index ID
        :return: True if truncated
        """
        return self.truncate_by_db_table("idx_ctx", store_id, idx)

    def truncate_external(self, store_id: str = None, idx: str = None) -> bool:
        """
        Truncate external table in database

        :param store_id: store ID
        :param idx: index ID
        :return: True if truncated
        """
        return self.truncate_by_db_table("idx_external", store_id, idx)

    def truncate_by_db_table(self, tbl: str, store_id: str = None, idx: str = None) -> bool:
        """
        Truncate external table in database

        :param tbl: table name
        :param store_id: store ID
        :param idx: index ID
        :return: True if truncated
        """
        db = self.window.core.db.get_db()
        query = "DELETE FROM " + tbl
        params = {}
        if store_id and idx:
            query += " WHERE store = :store_id AND idx = :idx"
            params = {"store_id": store_id, "idx": idx}
        elif store_id:
            query += " WHERE store = :store_id"
            params = {"store_id": store_id}
        elif idx:
            query += " WHERE idx = :idx"
            params = {"idx": idx}
        with db.begin() as conn:
            conn.execute(
                text(query).bindparams(**params))
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.27 04:00:00                  #
# ================================================== #

import datetime
from packaging.version import Version

from pygpt_net.item.index import IndexItem
from pygpt_net.provider.core.index.base import BaseProvider
from pygpt_net.provider.core.index.json_file import JsonFileProvider
from pygpt_net.provider.core.index.db_sqlite import DbSqliteProvider
from pygpt_net.provider.vector_stores import Storage

from .indexing import Indexing
from .llm import Llm
from .chat import Chat

from .types.ctx import Ctx
from .types.external import External
from .types.files import Files


class Idx:
    def __init__(self, window=None):
        """
        Indexer core

        :param window: Window instance
        """
        self.window = window
        self.indexing = Indexing(window)
        self.llm = Llm(window)
        self.storage = Storage(window)
        self.chat = Chat(window, self.storage)
        self.providers = {
            "json_file": JsonFileProvider(window),  # only for patching
            "db_sqlite": DbSqliteProvider(window),
        }
        self.provider = "db_sqlite"
        self.items = {}
        self.initialized = False

        # internal types
        self.ctx = Ctx(window, self.get_provider())
        self.external = External(window, self.get_provider())
        self.files = Files(window, self.get_provider())

    def install(self):
        """Install provider data"""
        self.get_provider().install()

    def patch(self, app_version: Version) -> bool:
        """
        Patch provider data

        :param app_version: app version
        :return: True if success
        """
        return self.get_provider().patch(app_version)

    def init(self):
        """Initialize indexes"""
        if not self.initialized:
            self.load()
            self.initialized = True

    def get_current_store(self) -> str:
        """
        Get current vector store name/ID

        :return: vector store name
        """
        return self.window.core.config.get('llama.idx.storage')

    def get_provider(self) -> BaseProvider:
        """
        Get provider instance

        :return: provider instance
        """
        return self.providers.get(self.provider)

    def store_index(self, idx: str = "base"):
        """
        Store (persist) index data

        :param idx: index name/ID
        """
        self.storage.store(idx)

    def remove_index(self, idx: str = "base", truncate: bool = False) -> bool:
        """
        Truncate index

        :param idx: index name
        :param truncate: truncate index (remove all data)
        :return: True if success
        """
        # get current store
        store = self.get_current_store()

        # clear db data
        self.ctx.truncate(store, idx)
        self.files.truncate(store, idx)

        # clear ctx data indexed status
        self.window.core.ctx.idx.truncate_indexed(store, idx)

        # clear storage, remove index
        if truncate:
            return self.storage.truncate(idx)
        else:
            return self.storage.remove(idx)

    def index_files(
            self,
            idx: str = "base",
            path: str = None
    ) -> (dict, list):
        """
        Index file or directory of files

        :param idx: index name
        :param path: path to file or directory
        :return: dict with indexed files (path -> id), list with errors
        """
        context = self.llm.get_service_context()
        index = self.storage.get(
            id=idx,
            service_context=context,
        )  # get or create index
        files, errors = self.indexing.index_files(
            idx=idx,
            index=index,
            path=path,
        )  # index files
        if len(files) > 0:
            self.storage.store(
                id=idx,
                index=index,
            )  # store index
        return files, errors

    def index_db_by_meta_id(
            self,
            idx: str = "base",
            id: int = 0,
            from_ts: int = 0
    ) -> (int, list):
        """
        Index records from db by meta id

        :param idx: index name
        :param id: CtxMeta id
        :param from_ts: timestamp from
        :return: num of indexed files, list with errors
        """
        context = self.llm.get_service_context()
        index = self.storage.get(
            id=idx,
            service_context=context,
        )  # get or create index
        num, errors = self.indexing.index_db_by_meta_id(
            idx=idx,
            index=index,
            id=id,
            from_ts=from_ts,
        )  # index db records
        if num > 0:
            self.storage.store(
                id=idx,
                index=index,
            )  # store index
        return num, errors

    def index_db_from_updated_ts(
            self,
            idx: str = "base",
            from_ts: int = 0
    ) -> (int, list):
        """
        Index records from db by meta id

        :param idx: index name
        :param from_ts: timestamp from
        :return: num of indexed files, list with errors
        """
        context = self.llm.get_service_context()
        index = self.storage.get(
            id=idx,
            service_context=context,
        )  # get or create index
        num, errors = self.indexing.index_db_from_updated_ts(
            idx=idx,
            index=index,
            from_ts=from_ts,
        )  # index db records
        if num > 0:
            self.storage.store(
                id=idx,
                index=index,
            )  # store index
        return num, errors

    def index_urls(
            self,
            idx: str = "base",
            urls: list = None,
            type: str = "webpage",
            extra_args: dict = None
    ) -> (dict, list):
        """
        Index URLs

        :param idx: index name
        :param urls: list of urls
        :param type: type of url
        :param extra_args: extra args
        :return: num of indexed, list with errors
        """
        context = self.llm.get_service_context()
        index = self.storage.get(
            id=idx,
            service_context=context,
        )  # get or create index
        n, errors = self.indexing.index_urls(
            idx=idx,
            index=index,
            urls=urls,
            type=type,
            extra_args=extra_args,
        )  # index urls
        if n > 0:
            self.storage.store(
                id=idx,
                index=index,
            )  # store index
        return n, errors

    def get_idx_data(self, idx: str = None) -> dict:
        """
        Get indexed files data

        :param idx: index name
        :return: indexed files data (idx -> items)
        """
        indexes = {}
        store_id = self.get_current_store()
        if idx is not None:
            if store_id in self.items and idx in self.items[store_id]:
                indexes[idx] = self.items[store_id][idx].items
        else:
            # all indexes
            if store_id in self.items:
                for idx in self.items[store_id]:
                    indexes[idx] = self.items[store_id][idx].items
        return indexes

    def get_by_idx(self, idx: int) -> str:
        """
        Return idx by list index

        :param idx: idx on list
        :return: idx name
        """
        items = self.window.core.config.get('llama.idx.list')
        if items is not None:
            if idx < len(items):
                return items[idx]['id']

    def get_idx_by_name(self, name: str) -> int:
        """
        Return idx on list by name

        :param name: idx name
        :return: idx on list
        """
        items = self.window.core.config.get('llama.idx.list')
        if items is not None:
            for idx, item in enumerate(items):
                if item['id'] == name:
                    return idx

    def get_default_idx(self) -> str:
        """
        Return default idx

        :return: idx name
        """
        if len(self.items) > 0:
            return self.get_by_idx(0)

    def has(self, idx: str) -> bool:
        """
        Check if index exists

        :param idx: index name
        :return: True if index exists
        """
        store_id = self.get_current_store()
        if store_id in self.items:
            return idx in self.items[store_id]
        return False

    def get(self, idx: str) -> IndexItem:
        """
        Return index data from current storage

        :param idx: index name
        :return: IndexItem object
        """
        store_id = self.get_current_store()
        if store_id in self.items and idx in self.items[store_id]:
            return self.items[store_id][idx]

    def get_all(self) -> dict:
        """
        Return all indexes in current store

        :return: all indexes
        """
        store_id = self.get_current_store()
        if store_id in self.items:
            return self.items[store_id]
        return {}

    def append(self, idx: str, files: dict):
        """
        Append indexed files to index

        :param idx: index name
        :param files: dict of indexed files
        """
        # create store if not exists
        store_id = self.get_current_store()
        if store_id not in self.items:
            self.items[store_id] = {}

        # create index if not exists
        if idx not in self.items[store_id]:
            self.items[store_id][idx] = IndexItem()
            self.items[store_id][idx].id = idx
            self.items[store_id][idx].name = idx  # use index id as name

        # append indexed files
        for path in files:
            doc_id = files[path]
            file_id = self.files.get_id(path)
            ts = int(datetime.datetime.now().timestamp())
            if file_id not in self.items[store_id][idx].items:
                id = self.files.append(
                    store_id=store_id,
                    idx=idx,
                    file_id=file_id,
                    path=path,
                    doc_id=doc_id,
                )
                if id is not None:
                    self.items[store_id][idx].items[file_id] = {
                        "id": doc_id,
                        "db_id": id,  # DB id
                        "path": path,
                        "indexed_ts": ts,
                    }
            else:
                # update indexed timestamp only
                self.files.update(
                    id=self.items[store_id][idx].items[file_id]["db_id"],  # DB id
                    doc_id=doc_id,
                    ts=ts,
                )
                self.items[store_id][idx].items[file_id]["id"] = doc_id
                self.items[store_id][idx].items[file_id]["indexed_ts"] = ts

    def remove_doc(self, idx: str, doc_id: str):
        """
        Remove document from index

        :param idx: index name (id)
        :param doc_id: document ID (in storage)
        """
        self.llm.get_service_context()  # init environment only (ENV API keys, etc.)
        if self.storage.remove_document(idx, doc_id):
            self.indexing.log("Removed document from index: " + idx + " - " + doc_id)

    def remove_file(self, idx: str, file: str):
        """
        Remove file from index

        :param idx: index name
        :param file: file ID
        """
        self.llm.get_service_context()  # init environment only (ENV API keys, etc.)
        store_id = self.get_current_store()
        if store_id in self.items and idx in self.items[store_id]:
            if file in self.items[store_id][idx].items:
                # remove from storage
                doc_id = self.items[store_id][idx].items[file]["id"]
                self.storage.remove_document(
                    id=idx,
                    doc_id=doc_id,
                )
                # remove from index data and db
                del self.items[store_id][idx].items[file]
                self.files.remove(store_id, idx, doc_id)

    def load(self):
        """Load indexes and indexed items"""
        store_ids = self.storage.get_ids()
        for store_id in store_ids:
            self.items[store_id] = self.get_provider().load(store_id)
            # replace workdir placeholder with current workdir
            for idx in self.items[store_id]:
                for id in self.items[store_id][idx].items:
                    file = self.items[store_id][idx].items[id]
                    if 'path' in file and file['path'] is not None:
                        self.items[store_id][idx].items[id]['path'] = \
                            self.window.core.filesystem.to_workdir(file['path'])

    def sync(self):
        """Sync idx items from config"""
        items = self.window.core.config.get('llama.idx.list')
        store_id = self.get_current_store()
        if items is not None:
            if store_id not in self.items:
                self.items[store_id] = {}
            for item in items:
                idx = item['id']
                if idx not in self.items[store_id]:
                    self.items[store_id][idx] = IndexItem()
                    self.items[store_id][idx].id = idx
                    self.items[store_id][idx].name = idx
                    self.items[store_id][idx].store = store_id
                else:
                    self.items[store_id][idx].id = idx
                    self.items[store_id][idx].name = idx

    def clear(self, idx: str):
        """
        Clear index items

        :param idx: index name/id
        """
        store_id = self.get_current_store()
        if store_id in self.items and idx in self.items[store_id]:
            self.items[store_id][idx].items = {}
        self.get_provider().truncate(store_id, idx)

    def get_version(self) -> str:
        """
        Get provider config version

        :return: provider config version
        """
        return self.get_provider().get_version()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.26 22:00:00                  #
# ================================================== #

import mimetypes
import os.path
import shutil
import ssl
from urllib.request import Request, urlopen
from PySide6.QtCore import Slot

from pygpt_net.plugin.base import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    pass  # add custom signals here


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None

    @Slot()
    def run(self):
        msg = None
        for item in self.cmds:
            try:
                if item["cmd"] in self.plugin.allowed_cmds \
                        and self.plugin.is_cmd_allowed(item["cmd"]):
                    request = {"cmd": item["cmd"]}  # prepare request item for result

                    # save file
                    if item["cmd"] == "save_file":
                        try:
                            msg = "Saving file: {}".format(item["params"]['filename'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['filename'],
                            )
                            data = item["params"]['data']
                            with open(path, 'w', encoding="utf-8") as file:
                                file.write(data)
                                response = {
                                    "request": request,
                                    "result": "OK",
                                }
                                self.log("File saved: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # append to file
                    elif item["cmd"] == "append_file" and self.plugin.is_cmd_allowed("append_file"):
                        try:
                            msg = "Appending file: {}".format(item["params"]['filename'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['filename'],
                            )
                            data = item["params"]['data']
                            with open(path, 'a', encoding="utf-8") as file:
                                file.write(data)
                                response = {
                                    "request": request,
                                    "result": "OK",
                                }
                                self.log("File appended: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # read file
                    elif item["cmd"] == "read_file" and self.plugin.is_cmd_allowed("read_file"):
                        try:
                            msg = "Reading file: {}".format(item["params"]['filename'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['filename'],
                            )

                            # check if file exists
                            if os.path.exists(path):

                                # auto-index file using Llama-index
                                if self.plugin.get_option_value("auto_index") \
                                        or self.plugin.get_option_value("only_index"):
                                    idx_name = self.plugin.get_option_value("idx")
                                    files, errors = self.plugin.window.core.idx.index_files(
                                        idx_name,
                                        path,
                                    )

                                    # if only index, return response and continue
                                    if self.plugin.get_option_value("only_index"):
                                        data = {
                                            'num_indexed': len(files),
                                            'index_name': idx_name,
                                            'errors': errors,
                                            'path': path,
                                        }
                                        response = {
                                            "request": request,
                                            "result": data,
                                        }
                                        self.log("File read (index only): {}".format(path))
                                        self.response(response)
                                        continue

                                # read file as text
                                data = self.plugin.read_as_text(
                                    path,
                                    use_loaders=self.plugin.get_option_value("use_loaders"),
                                )
                                response = {
                                    "request": request,
                                    "result": data,
                                }
                                self.log("File read: {}".format(path))
                            else:
                                response = {
                                    "request": request,
                                    "result": "File not found",
                                }
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # delete file
                    elif item["cmd"] == "delete_file" and self.plugin.is_cmd_allowed("delete_file"):
                        try:
                            msg = "Deleting file: {}".format(item["params"]['filename'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['filename'],
                            )
                            if os.path.exists(path):
                                os.remove(path)
                                response = {"request": request, "result": "OK"}
                                self.log("File deleted: {}".format(path))
                            else:
                                response = {
                                    "request": request,
                                    "result": "File not found",
                                }
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # list files
                    elif item["cmd"] == "list_dir" and self.plugin.is_cmd_allowed("list_dir"):
                        try:
                            msg = "Listing directory: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['path'],
                            )
                            if os.path.exists(path):
                                files = os.listdir(path)
                                response = {
                                    "request": request,
                                    "result": files,
                                }
                                self.log("Files listed: {}".format(path))
                                self.log("Result: {}".format(files))
                            else:
                                response = {
                                    "request": request,
                                    "result": "Directory not found",
                                }
                                self.log("Directory not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # mkdir
                    elif item["cmd"] == "mkdir" and self.plugin.is_cmd_allowed("mkdir"):
                        try:
                            msg = "Creating directory: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['path'],
                            )
                            if not os.path.exists(path):
                                os.makedirs(path)
                                response = {
                                    "request": request,
                                    "result": "OK",
                                }
                                self.log("Directory created: {}".format(path))
                            else:
                                response = {
                                    "request": request,
                                    "result": "Directory already exists",
                                }
                                self.log("Directory already exists: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # rmdir
                    elif item["cmd"] == "rmdir" and self.plugin.is_cmd_allowed("rmdir"):
                        try:
                            msg = "Deleting directory: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['path'],
                            )
                            if os.path.exists(path):
                                shutil.rmtree(path)
                                response = {
                                    "request": request,
                                    "result": "OK",
                                }
                                self.log("Directory deleted: {}".format(path))
                            else:
                                response = {
                                    "request": request,
                                    "result": "Directory not found",
                                }
                                self.log("Directory not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # download
                    elif item["cmd"] == "download_file" and self.plugin.is_cmd_allowed("download_file"):
                        try:
                            dst = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['dst'],
                            )
                            msg = "Downloading file: {} into {}".format(item["params"]['src'], dst)
                            self.log(msg)

                            # Check if src is URL
                            if item["params"]['src'].startswith("http"):
                                src = item["params"]['src']
                                # Download file from URL
                                req = Request(
                                    url=src,
                                    headers={'User-Agent': 'Mozilla/5.0'},
                                )
                                context = ssl.create_default_context()
                                context.check_hostname = False
                                context.verify_mode = ssl.CERT_NONE
                                with urlopen(
                                        req,
                                        context=context,
                                        timeout=5) as response, \
                                        open(dst, 'wb') as out_file:
                                    shutil.copyfileobj(response, out_file)

                                size = os.path.getsize(dst)
                            else:
                                # Handle local file paths
                                src = os.path.join(
                                    self.plugin.window.core.config.get_user_dir('data'),
                                    item["params"]['src'],
                                )

                                # Copy local file
                                with open(src, 'rb') as in_file, open(dst, 'wb') as out_file:
                                    shutil.copyfileobj(in_file, out_file)
                                size = os.path.getsize(dst)

                            # handle result
                            response = {
                                "request": request,
                                "result": {
                                    "result": "OK",
                                    "size_bytes": size,
                                    "size_human": self.get_human_readable_size(size),
                                }
                            }
                            self.log("File downloaded: {} into {}".format(src, dst))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # copy file
                    elif item["cmd"] == "copy_file" and self.plugin.is_cmd_allowed("copy_file"):
                        try:
                            msg = "Copying file: {} into {}".format(item["params"]['src'], item["params"]['dst'])
                            self.log(msg)
                            dst = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['dst'],
                            )
                            src = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['src'],
                            )
                            shutil.copyfile(src, dst)
                            response = {
                                "request": request,
                                "result": "OK",
                            }
                            self.log("File copied: {} into {}".format(src, dst))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # copy dir
                    elif item["cmd"] == "copy_dir" and self.plugin.is_cmd_allowed("copy_dir"):
                        try:
                            msg = "Copying directory: {} into {}".format(item["params"]['src'],
                                                                         item["params"]['dst'])
                            self.log(msg)
                            dst = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['dst'],
                            )
                            src = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['src'],
                            )
                            shutil.copytree(src, dst)
                            response = {
                                "request": request,
                                "result": "OK",
                            }
                            self.log("Directory copied: {} into {}".format(src, dst))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # move
                    elif item["cmd"] == "move" and self.plugin.is_cmd_allowed("move"):
                        try:
                            msg = "Moving: {} into {}".format(item["params"]['src'], item["params"]['dst'])
                            self.log(msg)
                            dst = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['dst'],
                            )
                            src = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['src'],
                            )
                            shutil.move(src, dst)
                            response = {
                                "request": request,
                                "result": "OK",
                            }
                            self.log("Moved: {} into {}".format(src, dst))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # is dir
                    elif item["cmd"] == "is_dir" and self.plugin.is_cmd_allowed("is_dir"):
                        try:
                            msg = "Checking if directory exists: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['path'],
                            )
                            if os.path.isdir(path):
                                response = {
                                    "request": request,
                                    "result": "OK",
                                }
                                self.log("Directory exists: {}".format(path))
                            else:
                                response = {"request": request, "result": "Directory not found"}
                                self.log("Directory not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # is file
                    elif item["cmd"] == "is_file" and self.plugin.is_cmd_allowed("is_file"):
                        try:
                            msg = "Checking if file exists: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['path'],
                            )
                            if os.path.isfile(path):
                                response = {
                                    "request": request,
                                    "result": "OK",
                                }
                                self.log("File exists: {}".format(path))
                            else:
                                response = {
                                    "request": request,
                                    "result": "File not found",
                                }
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # file exists
                    elif item["cmd"] == "file_exists" and self.plugin.is_cmd_allowed("file_exists"):
                        try:
                            msg = "Checking if path exists: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['path'],
                            )
                            if os.path.exists(path):
                                response = {
                                    "request": request,
                                    "result": "OK",
                                }
                                self.log("Path exists: {}".format(path))
                            else:
                                response = {
                                    "request": request,
                                    "result": "File or directory not found",
                                }
                                self.log("Path not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # file size
                    elif item["cmd"] == "file_size" and self.plugin.is_cmd_allowed("file_size"):
                        try:
                            msg = "Checking file size: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['path'],
                            )
                            if os.path.exists(path):
                                size = os.path.getsize(path)
                                response = {
                                    "request": request,
                                    "result": {
                                        'size_bytes': size,
                                        'size_human': self.plugin.human_readable_size(size),
                                    },
                                }
                                self.log("File size: {}".format(size))
                            else:
                                response = {
                                    "request": request,
                                    "result": "File not found",
                                }
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # file info
                    elif item["cmd"] == "file_info" and self.plugin.is_cmd_allowed("file_info"):
                        try:
                            msg = "Checking file info: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['path'],
                            )
                            if os.path.exists(path):
                                size = os.path.getsize(path)
                                data = {
                                    "size": size,
                                    "size_human": self.get_human_readable_size(size),
                                    'mime_type': mimetypes.guess_type(path)[0] or 'application/octet-stream',
                                    "last_access": os.path.getatime(path),
                                    "last_modification": os.path.getmtime(path),
                                    "creation_time": os.path.getctime(path),
                                    "is_dir": os.path.isdir(path),
                                    "is_file": os.path.isfile(path),
                                    "is_link": os.path.islink(path),
                                    "is_mount": os.path.ismount(path),
                                    'stat': os.stat(path),
                                }
                                response = {
                                    "request": request,
                                    "result": data,
                                }
                                self.log("File info: {}".format(data))
                            else:
                                response = {
                                    "request": request,
                                    "result": "File not found",
                                }
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # cwd
                    elif item["cmd"] == "cwd" and self.plugin.is_cmd_allowed("cwd"):
                        try:
                            msg = "Getting CWD: {}".format(self.plugin.window.core.config.get_user_dir('data'))
                            self.log(msg)
                            response = {
                                "request": request,
                                "result": self.plugin.window.core.config.get_user_dir('data'),
                            }
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # get file as attachment
                    elif item["cmd"] == "send_file" and self.plugin.is_cmd_allowed("send_file"):
                        try:
                            msg = "Adding attachment: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['path'],
                            )

                            if os.path.exists(path):
                                # make attachment
                                mode = self.plugin.window.core.config.get('mode')
                                title = os.path.basename(path)
                                self.plugin.window.core.attachments.new(mode, title, path, False)
                                self.plugin.window.core.attachments.save()
                                self.plugin.window.controller.attachment.update()
                                response = {
                                    "request": request,
                                    "result": "Sending attachment: {}".format(title),
                                }
                                self.log("Added attachment: {}".format(path))
                            else:
                                response = {
                                    "request": request,
                                    "result": "File not found",
                                }
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

                    # index file or directory
                    elif item["cmd"] == "file_index" and self.plugin.is_cmd_allowed("file_index"):
                        try:
                            msg = "Indexing path: {}".format(item["params"]['path'])
                            self.log(msg)
                            path = os.path.join(
                                self.plugin.window.core.config.get_user_dir('data'),
                                item["params"]['path'],
                            )
                            if os.path.exists(path):
                                idx_name = self.plugin.get_option_value("idx")
                                # index path using Llama-index
                                files, errors = self.plugin.window.core.idx.index_files(
                                    idx_name,
                                    path,
                                )
                                data = {
                                    'num_indexed': len(files),
                                    'index_name': idx_name,
                                    'errors': errors,
                                    'path': path,
                                }
                                response = {
                                    "request": request,
                                    "result": data,
                                }
                            else:
                                response = {
                                    "request": request,
                                    "result": "File or directory not found",
                                }
                                self.log("File not found: {}".format(path))
                        except Exception as e:
                            response = {
                                "request": request,
                                "result": "Error: {}".format(e),
                            }
                            self.error(e)
                            self.log("Error: {}".format(e))
                        self.response(response)

            except Exception as e:
                self.response(
                    {
                        "request": item,
                        "result": "Error: {}".format(e),
                    }
                )
                self.error(e)
                self.log("Error: {}".format(e))

        if msg is not None:
            self.status(msg)

        # update explorer view after file operations
        self.plugin.window.controller.files.update_explorer()

    def get_human_readable_size(self, size, decimal_places=2):
        """
        Return a human-readable file size.

        :param size: file size in bytes
        :param decimal_places: number of decimal places
        :return: human-readable file size
        """
        for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
            if size < 1024.0:
                break
            size /= 1024.0
        return f"{size:.{decimal_places}f} {unit}"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.27 04:00:00                  #
# ================================================== #

import datetime
import os


class IndexesDebug:
    def __init__(self, window=None):
        """
        Indexes debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'indexes'

    def update(self):
        """Update debug window."""
        self.window.core.debug.begin(self.id)
        self.window.core.debug.add(self.id, 'Current storage:', str(self.window.core.idx.get_current_store()))
        self.window.core.debug.add(self.id, 'Current idx:', str(self.window.controller.idx.current_idx))

        for store in self.window.core.idx.items:
            name = "Items (files): " + store
            indexes_data = {}
            for idx in self.window.core.idx.items[store]:
                indexes_data[idx] = len(self.window.core.idx.items[store][idx].items)
            if len(indexes_data) > 0:
                self.window.core.debug.add(self.id, name, str(indexes_data))

        self.window.core.debug.add(self.id, 'Storage (idx):',
                                   str(list(self.window.core.idx.storage.indexes.keys())))
        self.window.core.debug.add(self.id, 'Storage (storage):',
                                   str(list(self.window.core.idx.storage.storages.keys())))

        # loaders
        self.window.core.debug.add(self.id, 'Offline loaders [files]:',
                                   str(list(self.window.core.idx.indexing.loaders["file"].keys())))
        self.window.core.debug.add(self.id, 'Offline loaders [web]:',
                                   str(list(self.window.core.idx.indexing.loaders["web"].keys())))
        self.window.core.debug.add(self.id, 'External instructions [web]:',
                                   str(self.window.core.idx.indexing.external_instructions))

        # indexes
        indexes = self.window.core.idx.get_all()
        for key in list(indexes):
            path = os.path.join(self.window.core.config.get_user_dir('idx'), key)
            idx = indexes[key]
            self.window.core.debug.add(self.id, '----', '')
            idx_data = {
                'id': idx.id,
                'store': idx.store,
                'path': path,
                'items': len(idx.items),
            }
            self.window.core.debug.add(self.id, 'IDX: [' + str(key) + ']', str(idx_data))

            # files
            items = idx.items
            for item_id in items:
                item = items[item_id]
                indexed_dt = datetime.datetime.fromtimestamp(item['indexed_ts'])
                data = {
                    'id': item['id'],
                    'path': item['path'],
                    'indexed_at': str(item['indexed_ts']) + ' (' + str(indexed_dt) + ')',
                    'file_id': item_id,
                }
                self.window.core.debug.add(self.id, item_id, str(data))

        self.window.core.debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.27 04:00:00                  #
# ================================================== #

from .base import BaseProvider


class OpenAIWhisper(BaseProvider):
    def __init__(self, *args, **kwargs):
        """
        OpenAI Whisper provider (via API, not local model)

        :param args: args
        :param kwargs: kwargs
        """
        super(OpenAIWhisper, self).__init__(*args, **kwargs)
        self.plugin = kwargs.get("plugin")
        self.id = "openai_whisper"
        self.name = "OpenAI Whisper"

    def init_options(self):
        """Initialize options"""
        self.plugin.add_option(
            "whisper_model",
            type="text",
            value="whisper-1",
            label="Model",
            tab="openai_whisper",
            description="Specify model, default: whisper-1",
        )

    def transcribe(self, path: str) -> str:
        """
        Audio to text transcription

        :param path: path to audio file to transcribe
        :return: transcribed text
        """
        client = self.plugin.window.core.gpt.get_client()
        with open(path, "rb") as audio_file:
            return client.audio.transcriptions.create(
                model=self.plugin.get_option_value('whisper_model'),
                file=audio_file,
                response_format="text",
            )

    def is_configured(self) -> bool:
        """
        Check if provider is configured

        :return: True if configured, False otherwise
        """
        api_key = self.plugin.window.core.config.get("api_key")
        return api_key is not None and api_key != ""

    def get_config_message(self) -> str:
        """
        Return message to display when provider is not configured

        :return: message
        """
        return "OpenAI API key is not set yet. Please configure it in settings."



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.21 10:00:00                  #
# ================================================== #


class Common:
    def __init__(self, window=None):
        """
        Idx common controller

        :param window: Window instance
        """
        self.window = window

    def setup(self):
        """Set up UI"""
        # raw query
        if self.window.core.config.get('llama.idx.raw'):
            self.window.ui.config['global']['llama.idx.raw'].setChecked(True)
        else:
            self.window.ui.config['global']['llama.idx.raw'].setChecked(False)

    def enable_raw(self):
        """Enable raw query"""
        self.window.core.config.set('llama.idx.raw', True)
        self.window.core.config.save()

    def disable_raw(self):
        """Disable raw query"""
        self.window.core.config.set('llama.idx.raw', False)
        self.window.core.config.save()

    def toggle_raw(self, state: bool):
        """
        Toggle raw query

        :param state: state of checkbox
        """
        if not state:
            self.disable_raw()
        else:
            self.enable_raw()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.26 22:00:00                  #
# ================================================== #

class ContextDebug:
    def __init__(self, window=None):
        """
        Context debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'context'

    def update(self):
        """Update debug window"""
        self.window.core.debug.begin(self.id)
        self.window.core.debug.add(self.id, 'current (id)', str(self.window.core.ctx.current))
        self.window.core.debug.add(self.id, 'len(meta)', len(self.window.core.ctx.meta))
        self.window.core.debug.add(self.id, 'len(items)', len(self.window.core.ctx.items))
        self.window.core.debug.add(self.id, 'assistant', str(self.window.core.ctx.assistant))
        self.window.core.debug.add(self.id, 'mode', str(self.window.core.ctx.mode))
        self.window.core.debug.add(self.id, 'model', str(self.window.core.ctx.model))
        self.window.core.debug.add(self.id, 'preset', str(self.window.core.ctx.preset))
        self.window.core.debug.add(self.id, 'run', str(self.window.core.ctx.run))
        self.window.core.debug.add(self.id, 'status', str(self.window.core.ctx.status))
        self.window.core.debug.add(self.id, 'thread', str(self.window.core.ctx.thread))
        self.window.core.debug.add(self.id, 'last_mode', str(self.window.core.ctx.last_mode))
        self.window.core.debug.add(self.id, 'last_model', str(self.window.core.ctx.last_model))
        self.window.core.debug.add(self.id, 'search_string', str(self.window.core.ctx.search_string))
        self.window.core.debug.add(self.id, 'filters', str(self.window.core.ctx.filters))
        self.window.core.debug.add(self.id, 'sys_prompt (tmp)', str(self.window.core.ctx.current_sys_prompt))
        self.window.core.debug.add(self.id, 'allowed_modes', str(self.window.core.ctx.allowed_modes))

        current = None
        if self.window.core.ctx.current is not None:
            if self.window.core.ctx.current in self.window.core.ctx.meta:
                current = self.window.core.ctx.meta[self.window.core.ctx.current]
            if current is not None:
                data = {
                    "id": current.id,
                    "external_id": current.external_id,
                    "uuid": current.uuid,
                    "name": current.name,
                    "date": current.date,
                    "created": current.created,
                    "updated": current.updated,
                    "indexed": current.indexed,
                    "indexes": current.indexes,
                    "mode": current.mode,
                    "model": current.model,
                    "last_mode": current.last_mode,
                    "last_model": current.last_model,
                    "thread": current.thread,
                    "assistant": current.assistant,
                    "preset": current.preset,
                    "run": current.run,
                    "status": current.status,
                    "label": current.label,
                    "initialized": current.initialized,
                    "deleted": current.deleted,
                    "important": current.important,
                    "archived": current.archived,
                    "extra": current.extra,
                }
                self.window.core.debug.add(self.id, '*** (current)', str(data))

        i = 0
        self.window.core.debug.add(self.id, 'items[]', '')
        for item in self.window.core.ctx.items:
            data = {
                "id": item.id,
                "meta_id": item.meta_id,
                "external_id": item.external_id,
                "cmds": item.cmds,
                "results": item.results,
                "urls": item.urls,
                "images": item.images,
                "files": item.files,
                "attachments": item.attachments,
                "reply": item.reply,
                "input": item.input,
                "output": item.output,
                "mode": item.mode,
                "model": item.model,
                "thread": item.thread,
                "msg_id": item.msg_id,
                "run_id": item.run_id,
                "input_name": item.input_name,
                "output_name": item.output_name,
                "input_timestamp": item.input_timestamp,
                "output_timestamp": item.output_timestamp,
                "input_tokens": item.input_tokens,
                "output_tokens": item.output_tokens,
                "total_tokens": item.total_tokens,
                "extra": item.extra,
                "is_vision": item.is_vision,
                "idx": item.idx,
                "first": item.first,
                "tool_calls": item.tool_calls,
            }
            self.window.core.debug.add(self.id, str(item.id), str(data))
            i += 1

        self.window.core.debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.10 10:00:00                  #
# ================================================== #

from .field.checkbox import Checkbox
from .field.combo import Combo
from .field.dictionary import Dictionary
from .field.input import Input
from .field.slider import Slider
from .field.textarea import Textarea
from .placeholder import Placeholder


class Config:
    def __init__(self, window=None):
        """
        Configuration controller

        :param window: Window instance
        """
        self.window = window
        self.checkbox = Checkbox(window)
        self.combo = Combo(window)
        self.dictionary = Dictionary(window)
        self.input = Input(window)
        self.placeholder = Placeholder(window)
        self.slider = Slider(window)
        self.textarea = Textarea(window)

    def load_options(self, parent_id: str, options: dict):
        """
        Load options

        :param parent_id: Parent ID
        :param options: Options dict
        """
        for key in options:
            option = options[key]
            self.apply(parent_id, key, option)

    def apply(self, parent_id: str, key: str, option: dict):
        """
        Apply option to field handler based on type

        :param parent_id: Parent ID
        :param key: Option key
        :param option: Option dict
        """
        if option['type'] == 'int' or option['type'] == 'float':
            if 'slider' in option and option['slider']:
                self.slider.apply(parent_id, key, option)
            else:
                self.input.apply(parent_id, key, option)
        elif option['type'] == 'text':
            self.input.apply(parent_id, key, option)
        elif option['type'] == 'textarea':
            self.textarea.apply(parent_id, key, option)
        elif option['type'] == 'bool':
            self.checkbox.apply(parent_id, key, option)
        elif option['type'] == 'dict':
            self.dictionary.apply(parent_id, key, option)
        elif option['type'] == 'combo':
            self.combo.apply(parent_id, key, option)

    def apply_value(self, parent_id: str, key: str, option: dict, value):
        """
        Apply value to option

        :param parent_id: Parent ID
        :param key: Option key
        :param option: Option dict
        :param value: Option value
        """
        option['value'] = value
        self.apply(parent_id, key, option)

    def get_value(self, parent_id: str, key: str, option: dict) -> any:
        """
        Get value from field handler based on type

        :param parent_id: Parent ID
        :param key: Option key
        :param option: Option dict
        :return: Option value
        """
        if option['type'] == 'int' or option['type'] == 'float':
            if 'slider' in option and option['slider']:
                return self.slider.get_value(parent_id, key, option)
            else:
                return self.input.get_value(parent_id, key, option)
        elif option['type'] == 'text':
            return self.input.get_value(parent_id, key, option)
        elif option['type'] == 'textarea':
            return self.textarea.get_value(parent_id, key, option)
        elif option['type'] == 'bool':
            return self.checkbox.get_value(parent_id, key, option)
        elif option['type'] == 'dict':
            return self.dictionary.get_value(parent_id, key, option)
        elif option['type'] == 'combo':
            return self.combo.get_value(parent_id, key, option)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.25 01:00:00                  #
# ================================================== #
import datetime

from PySide6 import QtWidgets, QtCore, QtGui
from PySide6.QtGui import QAction, QIcon, QColor, QPixmap
from PySide6.QtWidgets import QMenu

from pygpt_net.ui.widget.lists.base import BaseList
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class ContextList(BaseList):
    def __init__(self, window=None, id=None):
        """
        Presets select menu

        :param window: main window
        :param id: input id
        """
        super(ContextList, self).__init__(window)
        self.window = window
        self.id = id

        self.doubleClicked.connect(self.dblclick)
        self.setItemDelegate(ImportantItemDelegate())

    def click(self, val):
        """
        Click event

        :param val: click event
        """
        self.window.controller.ctx.select_by_idx(val.row())

    def dblclick(self, val):
        """
        Double click event

        :param val: double click event
        """
        self.window.controller.ctx.select_by_idx(val.row())

    def contextMenuEvent(self, event):
        """
        Context menu event

        :param event: context menu event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        id = self.window.core.ctx.get_id_by_idx(idx)
        ctx_id = id
        ctx = self.window.core.ctx.get_meta_by_id(id)
        if ctx is None:
            return

        is_important = ctx.important

        actions = {}
        actions['rename'] = QAction(QIcon(":/icons/edit.svg"), trans('action.rename'), self)
        actions['rename'].triggered.connect(
            lambda: self.action_rename(event)
        )

        if is_important:
            actions['important'] = QAction(QIcon(":/icons/pin.svg"), trans('action.unpin'), self)
            actions['important'].triggered.connect(
                lambda: self.action_unpin(event)
            )
        else:
            actions['important'] = QAction(QIcon(":/icons/pin.svg"), trans('action.pin'), self)
            actions['important'].triggered.connect(
                lambda: self.action_pin(event)
            )

        actions['duplicate'] = QAction(QIcon(":/icons/copy.svg"), trans('action.duplicate'), self)
        actions['duplicate'].triggered.connect(
            lambda: self.action_duplicate(event)
        )

        actions['delete'] = QAction(QIcon(":/icons/delete.svg"), trans('action.delete'), self)
        actions['delete'].triggered.connect(
            lambda: self.action_delete(event)
        )

        actions['copy_id'] = QAction(QIcon(":/icons/copy.svg"), trans('action.ctx_copy_id') + " @" + str(id), self)
        actions['copy_id'].triggered.connect(
            lambda: self.action_copy_id(event)
        )

        menu = QMenu(self)
        menu.addAction(actions['rename'])
        menu.addAction(actions['duplicate'])
        menu.addAction(actions['important'])
        menu.addAction(actions['delete'])

        # set label menu
        set_label_menu = menu.addMenu(trans('calendar.day.label'))
        for status_id, status_info in self.window.controller.calendar.statuses.items():
            name = trans('calendar.day.' + status_info['label'])
            if status_id == 0:
                name = '-'
            color = status_info['color']
            pixmap = QPixmap(16, 16)
            pixmap.fill(color)
            icon = QIcon(pixmap)
            status_action = QAction(icon, name, self)
            status_action.triggered.connect(
                lambda checked=False, s_id=status_id: self.window.controller.ctx.set_label(idx, s_id)
            )
            set_label_menu.addAction(status_action)

        idx_menu = QMenu(trans('action.idx'), self)

        # indexes list
        idxs = self.window.core.config.get('llama.idx.list')
        store = self.window.core.idx.get_current_store()  # get current idx store provider
        if len(idxs) > 0:
            for index in idxs:
                id = index['id']
                name = index['name'] + " (" + index['id'] + ")"

                # add to index
                action = idx_menu.addAction("IDX: " + name)
                action.setIcon(QIcon(":/icons/search.svg"))
                action.triggered.connect(
                    lambda checked=False,
                           idx=idx,
                           index=id: self.action_idx(idx, index)
                )

                # remove from index
                if ctx.indexed is not None and ctx.indexed > 0:

                    # get list of indexes in which context is indexed
                    if store in ctx.indexes:
                        store_indexes = ctx.indexes[store]
                        for store_index in store_indexes:
                            action = idx_menu.addAction(trans("action.idx.remove") + ": " + store_index)
                            action.setIcon(QIcon(":/icons/delete.svg"))
                            action.triggered.connect(
                                lambda checked=False,
                                       store_index=store_index,
                                       ctx_id=ctx_id: self.action_idx_remove(store_index, ctx_id)  # by context meta id
                            )

            menu.addMenu(idx_menu)

        menu.addAction(actions['copy_id'])

        # show last indexed date if available
        if ctx.indexed is not None and ctx.indexed > 0:
            suffix = ""
            if ctx.updated > ctx.indexed:
                suffix = " *"
            dt = datetime.datetime.fromtimestamp(ctx.indexed).strftime("%Y-%m-%d %H:%M")
            action = QAction(QIcon(":/icons/clock.svg"), trans('action.ctx.indexed') + ": " + dt + suffix, self)
            action.setEnabled(False)  # disable action, only for info
            menu.addAction(action)

        if idx >= 0:
            self.window.controller.ctx.select_by_idx(item.row())
            menu.exec_(event.globalPos())

    def action_idx(self, ctx_idx: int, idx: int):
        """
        Index with llama context action handler

        :param ctx_idx: row idx in context list
        :param idx: index name
        """
        self.window.controller.idx.indexer.index_ctx_meta(ctx_idx, idx)

    def action_idx_remove(self, idx: str, meta_id: int):
        """
        Remove from index action handler

        :param idx: index id
        :param meta_id: meta id
        """
        self.window.controller.idx.indexer.index_ctx_meta_remove(idx, meta_id)

    def action_rename(self, event):
        """
        Rename action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.ctx.rename(idx)

    def action_pin(self, event):
        """
        Pin action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.ctx.set_important(idx, True)

    def action_unpin(self, event):
        """
        Unpin action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.ctx.set_important(idx, False)

    def action_duplicate(self, event):
        """
        Rename duplicate handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.ctx.common.duplicate_by_idx(idx)

    def action_important(self, event):
        """
        Set as important action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.ctx.set_important(idx)

    def action_delete(self, event):
        """
        Delete action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.ctx.delete(idx)

    def action_copy_id(self, event):
        """
        Copy ID tag action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.ctx.common.copy_id(idx)


class ImportantItemDelegate(QtWidgets.QStyledItemDelegate):
    """
    Label color delegate

    :param QtWidgets.QStyledItemDelegate: parent class
    """
    def paint(self, painter, option, index):
        super(ImportantItemDelegate, self).paint(painter, option, index)

        # pin (>= 10)
        if index.data(QtCore.Qt.ItemDataRole.UserRole) > 0:
            label = index.data(QtCore.Qt.ItemDataRole.UserRole)
            painter.save()

            if label >= 10:
                color = self.get_color_for_status(3)
                square_size = 3
                square_margin = 0
                square_rect = QtCore.QRect(
                    option.rect.left() + square_margin,
                    option.rect.top() + 2,
                    square_size,
                    square_size,
                )
                painter.setBrush(color)
                painter.setPen(
                    QtGui.QPen(
                        QtCore.Qt.black,
                        0.5,
                        QtCore.Qt.SolidLine,
                    )
                )
                painter.drawRect(square_rect)

                label = label - 10  # remove pin status

            # label (0-9)
            if label > 0:
                color = self.get_color_for_status(label)
                square_size = 5
                square_margin = 0
                square_rect = QtCore.QRect(
                    option.rect.left() + square_margin,
                    option.rect.center().y() - (square_size / 2) + 2,
                    square_size,
                    square_size,
                )
                painter.setBrush(color)
                painter.setPen(QtCore.Qt.NoPen)
                painter.drawRect(square_rect)

            painter.restore()

    def get_color_for_status(self, status: int) -> QColor:
        """
        Get color for status

        :param status: status id
        :return: color
        """
        statuses = {
            0: {'label': 'label.color.default', 'color': QColor(100, 100, 100)},
            1: {'label': 'label.color.red', 'color': QColor(255, 0, 0)},
            2: {'label': 'label.color.orange', 'color': QColor(255, 165, 0)},
            3: {'label': 'label.color.yellow', 'color': QColor(255, 255, 0)},
            4: {'label': 'label.color.green', 'color': QColor(0, 255, 0)},
            5: {'label': 'label.color.blue', 'color': QColor(0, 0, 255)},
            6: {'label': 'label.color.indigo', 'color': QColor(75, 0, 130)},
            7: {'label': 'label.color.violet', 'color': QColor(238, 130, 238)},
        }
        if status in statuses:
            return statuses[status]['color']
        else:
            return statuses[0]['color']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.22 10:00:00                  #
# ================================================== #

from pygpt_net.utils import trans


class Settings:
    def __init__(self, window=None):
        """
        Settings locale controller

        :param window: Window instance
        """
        self.window = window

    def apply(self):
        """Apply locale to settings dialog"""
        # load settings options if not loaded yet
        if not self.window.controller.settings.editor.initialized:
            self.window.controller.settings.editor.load_config_options(False)

        # update settings options labels
        for id in self.window.controller.settings.editor.options:
            option = self.window.controller.settings.editor.options[id]
            option_label = 'settings.{}.label'.format(id)  # TODO: check
            option_desc = 'settings.{}.desc'.format(id)  # TODO: check
            trans_key = '{}'.format(option['label'])

            # label
            if option['type'] == 'bool':
                if id in self.window.ui.config['config']:
                    self.window.ui.config['config'][id].box.setText(trans(trans_key))
            else:
                if option_label in self.window.ui.nodes:
                    self.window.ui.nodes[option_label].setText(trans(trans_key))

            # description
            if option_desc in self.window.ui.nodes:
                if 'description' in option \
                        and option['description'] is not None \
                        and option['description'].strip() != "":
                    trans_desc_key = '{}'.format(option['description'])
                    self.window.ui.nodes[option_desc].setText(trans(trans_desc_key))

        # update sections tabs
        sections = self.window.core.settings.get_sections()
        i = 0
        for section_id in sections.keys():
            key = 'settings.section.' + section_id.replace("-", "_")
            self.window.ui.tabs['settings.section'].setTabText(i, trans(key))
            i += 1

        # update sections list
        idx = self.window.ui.tabs['settings.section'].currentIndex()
        self.window.settings.refresh_list()
        self.window.controller.settings.set_by_tab(idx)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.21 01:00:00                  #
# ================================================== #

from PySide6.QtCore import QObject, Signal, QRunnable, Slot

from pygpt_net.item.ctx import CtxItem


class AssistantsWorker:
    def __init__(self, window=None):
        """
        Assistants Worker (async)

        :param window: Window instance
        """
        self.window = window

    @Slot(object, object)
    def handle_error(self, ctx: CtxItem, err: any):
        """
        Handle thread error signal

        :param ctx: context item
        :param err: error message
        """
        self.window.controller.assistant.threads.handle_run_error(ctx, err)

    @Slot(object, object)
    def handle_run_created(self, ctx: CtxItem, run):
        """
        Handle thread finished signal

        :param ctx: context item
        :param run
        """
        self.window.controller.assistant.threads.handle_run_created(ctx, run)

    def create_run(
            self,
            ctx: CtxItem,
            thread_id: str,
            assistant_id: str,
            system_prompt: str
    ):
        """
        Create assistant run

        :param ctx: context item
        :param thread_id: thread id
        :param assistant_id: assistant id
        :param system_prompt: system prompt
        """
        worker = Worker()
        worker.window = self.window
        worker.mode = "run_create"
        worker.ctx = ctx
        worker.thread_id = thread_id
        worker.assistant_id = assistant_id
        worker.system_prompt = system_prompt
        worker.signals.finished.connect(self.handle_run_created)
        worker.signals.error.connect(self.handle_error)
        self.window.threadpool.start(worker)

    def msg_send(
            self,
            ctx: CtxItem,
            thread_id: str,
            assistant_id: str,
            prompt: str,
            system_prompt: str
    ):
        """
        Send message to assistant thread

        :param ctx: context item
        :param thread_id: thread id
        :param assistant_id: assistant id
        :param prompt: prompt
        :param system_prompt: system prompt
        """
        worker = Worker()
        worker.window = self.window
        worker.mode = "msg_send"
        worker.ctx = ctx
        worker.thread_id = thread_id
        worker.assistant_id = assistant_id
        worker.prompt = prompt
        worker.system_prompt = system_prompt
        worker.signals.finished.connect(self.handle_run_created)
        worker.signals.error.connect(self.handle_error)
        self.window.threadpool.start(worker)

    def tools_submit(
            self,
            ctx: CtxItem,
            tools_outputs: list,
    ):
        """
        Send tools outputs to assistant thread

        :param ctx: context item
        :param tools_outputs: list of tools outputs
        """
        worker = Worker()
        worker.window = self.window
        worker.mode = "tools_submit"
        worker.ctx = ctx
        worker.tools_outputs = tools_outputs
        worker.signals.finished.connect(self.handle_run_created)
        worker.signals.error.connect(self.handle_error)
        self.window.threadpool.start(worker)


class WorkerSignals(QObject):
    finished = Signal(object, object)
    error = Signal(object, object)


class Worker(QRunnable):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.window = None
        self.mode = None
        self.thread_id = None
        self.assistant_id = None
        self.prompt = None
        self.system_prompt = None
        self.tools_outputs = None
        self.ctx = None

    @Slot()
    def run(self):
        """Assistants worker thread"""
        if self.mode == "run_create":
            self.run_create()
        elif self.mode == "msg_send":
            self.msg_send()
        elif self.mode == "tools_submit":
            self.tools_submit()

    def run_create(self) -> bool:
        """
        Create assistant run

        :return: result
        """
        try:
            run = self.window.core.gpt.assistants.run_create(
                self.thread_id,
                self.assistant_id,
                self.system_prompt,
            )
            if run is not None:
                self.signals.finished.emit(self.ctx, run)
                return True
        except Exception as e:
            self.signals.error.emit(self.ctx, e)
        return False

    def msg_send(self) -> bool:
        """
        Send message to assistant

        :return: result
        """
        try:
            response = self.window.core.gpt.assistants.msg_send(
                self.thread_id,
                self.prompt,
            )
            if response is not None:
                self.ctx.msg_id = response.id
                return self.run_create()
        except Exception as e:
            self.signals.error.emit(self.ctx, e)
        return False

    def tools_submit(self) -> bool:
        """
        Submit tools outputs to assistant

        :return: result
        """
        try:
            run = self.window.core.gpt.assistants.run_submit_tool(self.ctx, self.tools_outputs)
            if run is not None:
                self.ctx.run_id = run.id  # update run id
                self.signals.finished.emit(self.ctx, run)  # continue status check
                return True
        except Exception as e:
            self.signals.error.emit(self.ctx, e)
        return False



"""Read Microsoft Word files."""

from pathlib import Path
from typing import Dict, List, Optional

from llama_index.readers.base import BaseReader
from llama_index.readers.schema.base import Document


class WebPage(BaseReader):
    """Webpage base reader."""

    def load_data(self, **kwargs) -> List[Document]:
        """
        Read URL and return documents.

        :param kwargs: keyword arguments
        :return: list of documents
        """
        from llama_index.readers import BeautifulSoupWebReader

        url = kwargs.get("url")

        return BeautifulSoupWebReader().load_data([url])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.01.12 04:00:00                  #
# ================================================== #

import os
import shutil
import sys
import time

from packaging.version import parse as parse_version, Version

from pygpt_net.item.notepad import NotepadItem


class Patch:
    def __init__(self, window=None, provider=None):
        self.window = window
        self.provider = provider

    def execute(self, version: Version) -> bool:
        """
        Migrate to current app version

        :param version: current app version
        :return: True if migrated
        """
        # return
        # if old version is 2.0.59 or older and if json file exists
        path = os.path.join(self.window.core.config.path, 'notepad.json')
        if os.path.exists(path):
            self.provider.truncate()
            self.import_from_json()
            os.rename(path, path + ".old")  # rename notepad.json to notepad.json.old
            return True

    def import_from_json(self) -> bool:
        """
        Import notepads from JSON file

        :return: True if imported
        :rtype: bool
        """
        # use json provider to load old notepads
        provider = self.window.core.notepad.providers['json_file']
        provider.attach(self.window)

        print("[DB] Migrating into database storage...")
        print("[DB] Importing notepads from JSON files... this may take a while. Please wait...")
        i = 0
        notepads = provider.load_all()
        cols, _ = shutil.get_terminal_size()
        c = len(notepads)
        ts = int(time.time())
        for id in notepads:
            notepad = notepads[id]
            notepad.created = ts
            notepad.updated = ts
            notepad.idx = int(id)
            notepad.uuid = self.provider.create_id()

            line = "[DB] Importing notepad %s/%s" % (i + 1, c)
            print(f"{line:<{cols}}", end='\r')
            sys.stdout.flush()

            self.import_notepad(notepad)
            i += 1

        print()  # new line
        if i > 0:
            print("[DB][DONE] Imported %s notepads." % i)
            return True

    def import_notepad(self, notepad: NotepadItem):
        """
        Import notepad from JSON file

        :param notepad: NotepadItem
        """
        notepad.id = None  # reset old ID to allow creating new
        self.provider.create(notepad)  # create new notepad and get new ID



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.15 01:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QPushButton, QHBoxLayout, QLabel, QVBoxLayout, QSplitter, QWidget, QSizePolicy

from pygpt_net.ui.base.config_dialog import BaseConfigDialog
from pygpt_net.ui.widget.dialog.editor import EditorDialog
from pygpt_net.utils import trans


class Assistant(BaseConfigDialog):
    def __init__(self, window=None, *args, **kwargs):
        super(Assistant, self).__init__(window, *args, **kwargs)
        """
        Assistant editor dialog

        :param window: Window instance
        """
        self.window = window
        self.id = "assistant"
        self.dialog_id = "assistants"

    def setup(self):
        """Setups assistant editor dialog"""
        self.window.ui.nodes['assistant.btn.save'] = QPushButton(trans("dialog.assistant.btn.save"))
        self.window.ui.nodes['assistant.btn.save'].clicked.connect(
            lambda: self.window.controller.assistant.editor.save()
        )
        self.window.ui.nodes['assistant.btn.save'].setAutoDefault(True)

        footer = QHBoxLayout()
        footer.addWidget(self.window.ui.nodes['assistant.btn.save'])

        # get option fields config
        fields = self.window.controller.assistant.editor.get_options()

        # build settings widgets
        widgets = self.build_widgets(self.id, fields)  # from base config dialog

        # apply settings widgets
        for key in widgets:
            self.window.ui.config[self.id][key] = widgets[key]

        # btn: add function
        self.window.ui.config[self.id]['tool.function'].add_btn.setText(trans('assistant.func.add'))
        # Empty params: {"type": "object", "properties": {}}

        # set tips
        self.window.ui.nodes['assistant.id_tip'] = QLabel(trans('assistant.new.id_tip'))
        self.window.ui.nodes['assistant.id_tip'].setMinimumHeight(40)

        self.window.ui.nodes['assistant.api.tip'] = QPushButton(trans('assistant.api.tip'))  # TODO: url btn
        self.window.ui.nodes['assistant.api.tip'].setAutoDefault(False)
        self.window.ui.nodes['assistant.api.tip'].setFlat(True)
        self.window.ui.nodes['assistant.api.tip'].setStyleSheet(
            "text-align: left; "
            "color: #fff; "
            "text-decoration: underline; "
            "text-transform: none;"
        )
        self.window.ui.nodes['assistant.api.tip'].setStyleSheet("text-transform: none;")
        self.window.ui.nodes['assistant.api.tip'].setCursor(Qt.PointingHandCursor)
        self.window.ui.nodes['assistant.api.tip'].clicked.connect(
            lambda: self.window.controller.assistant.goto_online()
        )

        # apply widgets to layouts
        options = {}
        for key in widgets:
            if fields[key]["type"] == 'text' or fields[key]["type"] == 'int' or fields[key]["type"] == 'float':
                options[key] = self.add_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'textarea':
                options[key] = self.add_row_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'bool':
                options[key] = self.add_raw_option(widgets[key], fields[key])
            elif fields[key]["type"] == 'dict':
                options[key] = self.add_row_option(widgets[key], fields[key])
                if key == "tool.function":
                    widgets[key].setMinimumHeight(150)
            elif fields[key]["type"] == 'combo':
                options[key] = self.add_option(widgets[key], fields[key])

        self.window.ui.nodes['assistant.tool.function.label'].setVisible(False)  # hide label
        self.window.ui.nodes['assistant.id_tip'].setAlignment(Qt.AlignCenter)

        options["tool.code_interpreter"].setAlignment(Qt.AlignCenter)
        options["tool.retrieval"].setAlignment(Qt.AlignCenter)

        rows = QVBoxLayout()
        rows.addLayout(options["id"])
        rows.addLayout(options["name"])
        rows.addLayout(options["description"])
        rows.addLayout(options["model"])
        rows.addWidget(self.window.ui.nodes['assistant.id_tip'])
        rows.addWidget(self.window.ui.nodes['assistant.api.tip'])
        rows.setContentsMargins(0, 0, 0, 0)

        rows_tools = QHBoxLayout()
        rows_tools.addLayout(options["tool.code_interpreter"])
        rows_tools.addLayout(options["tool.retrieval"])
        rows_tools.setContentsMargins(0, 0, 0, 0)

        rows.addLayout(rows_tools)
        rows.addStretch()

        widget_base = QWidget()
        widget_base.setLayout(rows)

        widget_base.setMinimumWidth(400)
        widget_base.setMaximumWidth(450)

        widget_prompt = QWidget()
        widget_prompt.setLayout(options["instructions"])
        widget_prompt.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        options["instructions"].setContentsMargins(0, 0, 0, 0)
        options["tool.function"].setContentsMargins(0, 0, 0, 0)

        widget_tools = QWidget()
        widget_tools.setLayout(options["tool.function"])
        widget_tools.setMinimumWidth(400)

        main = QHBoxLayout()
        main.addWidget(widget_base)
        main.addWidget(widget_tools)
        main.setContentsMargins(0, 0, 0, 0)

        widget_main = QWidget()
        widget_main.setLayout(main)

        self.window.ui.splitters['editor.assistant'] = QSplitter(Qt.Vertical)
        self.window.ui.splitters['editor.assistant'].addWidget(widget_main)
        self.window.ui.splitters['editor.assistant'].addWidget(widget_prompt)
        self.window.ui.splitters['editor.assistant'].setStretchFactor(0, 1)
        self.window.ui.splitters['editor.assistant'].setStretchFactor(1, 2)

        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.splitters['editor.assistant'])
        layout.addLayout(footer)

        self.window.ui.dialog['editor.' + self.dialog_id] = EditorDialog(self.window, self.dialog_id)
        self.window.ui.dialog['editor.' + self.dialog_id].setLayout(layout)
        self.window.ui.dialog['editor.' + self.dialog_id].setWindowTitle(trans('dialog.assistant'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.23 01:00:00                  #
# ================================================== #
import os
import shutil

from llama_index.indices.base import BaseIndex
from llama_index import (
    ServiceContext,
)


class BaseStore:
    def __init__(self, *args, **kwargs):
        """
        Base vector store provider

        :param args: args
        :param kwargs: kwargs
        """
        self.window = kwargs.get('window', None)
        self.id = None
        self.prefix = ""  # prefix for index directory
        self.indexes = {}

    def attach(self, window=None):
        """
        Attach window instance

        :param window: Window instance
        """
        self.window = window

    def get_path(self, id: str) -> str:
        """
        Get database path

        :param id: index name
        :return: database path
        """
        return os.path.join(
            self.window.core.config.get_user_dir('idx'),
            self.prefix + id,
        )

    def exists(self, id: str = None) -> bool:
        """
        Check if index with id exists

        :param id: index name
        :return: True if exists
        """
        path = self.get_path(id)
        return os.path.exists(path)

    def create(self, id: str):
        """
        Create empty index

        :param id: index name
        """
        pass

    def get(self, id: str, service_context: ServiceContext = None) -> BaseIndex:
        """
        Get index instance

        :param id: index name
        :param service_context: Service context
        :return: index instance
        """
        pass

    def store(self, id: str, index: BaseIndex = None):
        """
        Store/persist index

        :param id: index name
        :param index: index instance
        """
        pass

    def remove(self, id: str) -> bool:
        """
        Clear index

        :param id: index name
        :return: True if success
        """
        if id in self.indexes:
            self.indexes[id] = None
        path = self.get_path(id)
        if os.path.exists(path):
            shutil.rmtree(path)
        return True

    def truncate(self, id: str) -> bool:
        """
        Truncate index

        :param id: index name
        :return: True if success
        """
        return self.remove(id)

    def remove_document(self, id: str, doc_id: str) -> bool:
        """
        Remove document from index

        :param id: index name
        :param doc_id: document ID
        :return: True if success
        """
        index = self.get(id)
        index.delete(doc_id)
        self.store(
            id=id,
            index=index,
        )
        return True
